# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining `gaarf-simulator` CLI utility.

`gaarf-simulator` allows to simulate API resonse based on GAQL queries
and store results in local/remote storage.
"""

from __future__ import annotations

import argparse

import yaml

from gaarf import api_clients, simulation
from gaarf.cli import utils
from gaarf.io import reader, writer


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('query', nargs='+')
  parser.add_argument('-c', '--config', dest='gaarf_config', default=None)
  parser.add_argument(
    '-s', '--simulator-config', dest='simulator_config', default=None
  )
  parser.add_argument('--account', dest='customer_id', default='None')
  parser.add_argument('--output', dest='save', default='console')
  parser.add_argument('--input', dest='input', default='file')
  parser.add_argument(
    '--api-version',
    dest='api_version',
    default=api_clients.GOOGLE_ADS_API_VERSION,
  )
  parser.add_argument('--log', '--loglevel', dest='loglevel', default='info')
  parser.add_argument('--logger', dest='logger', default='local')
  parser.add_argument(
    '--customer-ids-query', dest='customer_ids_query', default=None
  )
  parser.add_argument(
    '--customer-ids-query-file', dest='customer_ids_query_file', default=None
  )
  parser.add_argument('--save-config', dest='save_config', action='store_true')
  parser.add_argument(
    '--no-save-config', dest='save_config', action='store_false'
  )
  parser.add_argument(
    '--config-destination', dest='save_config_dest', default='config.yaml'
  )
  parser.set_defaults(save_config=False)
  args = parser.parse_known_args()
  main_args = args[0]

  logger = utils.init_logging(
    loglevel=main_args.loglevel.upper(), logger_type=main_args.logger
  )

  config = utils.ConfigBuilder('gaarf').build(vars(args[0]), args[1])
  logger.debug('config: %s', config)

  writer_client = writer.WriterFactory().create_writer(
    config.output, **config.writer_params
  )
  if config.output == 'bq':
    _ = writer_client.create_or_get_dataset()
  reader_factory = reader.ReaderFactory()
  reader_client = reader_factory.create_reader(main_args.input)

  for query in main_args.query:
    if simulator_config := main_args.simulator_config:
      with open(simulator_config, encoding='utf-8') as f:
        simulator_config = yaml.safe_load(f)
      simulator_specification = simulation.SimulatorSpecification(
        **simulator_config
      )
    else:
      simulator_specification = simulation.SimulatorSpecification()
    logger.info('Simulating data for query %s', query)
    if report := simulation.simulate_data(
      reader_client.read(query),
      query,
      config.params,
      main_args.api_version,
      simulator_specification,
    ):
      writer_client.write(report, query)
    else:
      logger.info('Cannot simulate data for query %s', query)


if __name__ == '__main__':
  main()



from __future__ import annotations

import pytest

from gaarf.api_clients import GOOGLE_ADS_API_VERSION
from gaarf.query_editor import QuerySpecification
from gaarf.simulation import SimulatorSpecification, simulate_data


@pytest.fixture
def query():
  return """
        SELECT
            segments.date AS date,
            campaign.advertising_channel_sub_type AS campaign_type,
            campaign.id as campaign_id,
            campaign.name AS campaign_name,
            asset.youtube_video_asset.youtube_video_id AS video_id,
            asset.image_asset.full_size.url AS url,
            metrics.cost_micros AS cost,
            metrics.clicks AS clicks,
            metrics.conversions AS conversions
        FROM campaign
        """


class TestDefaultSimulation:
  @pytest.fixture
  def query_specification(self, query):
    return QuerySpecification(query).generate()

  @pytest.fixture
  def default_simulator_specification(self):
    return SimulatorSpecification(n_rows=3)

  @pytest.fixture
  def default_report(self, query, default_simulator_specification):
    return simulate_data(
      query_text=query,
      query_name='test',
      args=None,
      api_version=GOOGLE_ADS_API_VERSION,
      simulator_specification=default_simulator_specification,
    )

  def test_simulate_data_returns_same_n_rows_as_in_simulation_specification(
    self, default_report, default_simulator_specification
  ):
    assert len(default_report) == default_simulator_specification.n_rows

  def test_simulate_data_return_correct_number_of_columns(
    self, default_report, query_specification
  ):
    assert len(default_report.results[0]) == len(query_specification.fields)

  def test_simulate_data_return_correct_id_column(
    self, default_report, query_specification
  ):
    assert 1000000 <= default_report[0].campaign_id <= 1000010

  def test_simulate_data_return_correct_video_id_column(
    self, default_report, query_specification
  ):
    assert default_report[0].video_id == '4WXs3sKu41I'

  def test_simulate_data_return_correct_url_column(
    self, default_report, query_specification
  ):
    assert default_report[0].url == 'example.com'

  def test_simulate_data_return_correct_micros_column(
    self, default_report, query_specification
  ):
    assert 0 <= default_report[0].cost <= 1000 * 1e6

  def test_simulate_data_return_correct_int_column(
    self, default_report, query_specification
  ):
    assert 0 <= default_report[0].clicks <= 1000

  def test_simulate_data_return_correct_float_column(
    self, default_report, query_specification
  ):
    assert 0 <= default_report[0].conversions <= 1000


class TestCustomSimulation:
  @pytest.fixture
  def allowed_enums(self):
    return ['APP_CAMPAIGN', 'APP_CAMPAIGN_FOR_ENGAGEMENT']

  @pytest.fixture
  def replacements(self):
    return {'asset.youtube_video_asset.youtube_video_id': ['12345', '54321']}

  @pytest.fixture
  def custom_simulator_specification(self, allowed_enums, replacements):
    return SimulatorSpecification(
      allowed_enums={'campaign.advertising_channel_sub_type': allowed_enums},
      replacements=replacements,
    )

  @pytest.fixture
  def custom_report(self, query, custom_simulator_specification):
    return simulate_data(
      query_text=query,
      query_name='test',
      args=None,
      api_version=GOOGLE_ADS_API_VERSION,
      simulator_specification=custom_simulator_specification,
    )

  def test_simulate_data_returns_only_allowed_enums(
    self, custom_report, custom_simulator_specification, allowed_enums
  ):
    report_iterator = iter(custom_report)
    assert next(report_iterator).get('campaign_type') in allowed_enums
    assert next(report_iterator).get('campaign_type') in allowed_enums

  def test_simulate_data_return_correct_replacements(
    self, custom_report, custom_simulator_specification, replacements
  ):
    report_iterator = iter(custom_report)
    replacement = replacements.get('asset.youtube_video_asset.youtube_video_id')
    assert next(report_iterator).get('video_id') in replacement
    assert next(report_iterator).get('video_id') in replacement


class TestSimulationForBuiltInQuery:
  def test_simulate_data_returns_none_for_builtin_query(self):
    query = 'SELECT * FROM builtin.ocid_mapping'
    report = simulate_data(
      query_text=query,
      query_name='test',
      args=None,
      api_version=GOOGLE_ADS_API_VERSION,
      simulator_specification=SimulatorSpecification(),
    )
    assert report is None



import pytest

from gaarf.query_post_processor import PostProcessorMixin


@pytest.fixture
def template_query_with_if_else():
  return "SELECT field_one, {% if key == 'field_2' %} field_two {% else %} field_three {% endif %} FROM some_table"


@pytest.fixture
def template_query_with_for_loop():
  return 'SELECT field_one, {% for day in cohort_days %} {{day}} AS day_{{day}}, {% endfor %} FROM some_table'


@pytest.fixture
def macro_template_query_with_for_loop():
  return 'SELECT field_one, {field_2} AS field_two, {% for day in cohort_days %} {{day}} AS day_{{day}}, {% endfor %} FROM some_table'


def test_expand_jinja_if_template_parameters_are_provided_as_comma_separated_string(
  template_query_with_for_loop,
):
  expected_query = 'SELECT field_one, 1 AS day_1, 2 AS day_2, FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_for_loop, {'cohort_days': '1,2'}
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_if_template_parameters_are_provided_as_list(
  template_query_with_for_loop,
):
  expected_query = 'SELECT field_one, 1 AS day_1, 2 AS day_2, FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_for_loop, {'cohort_days': [1, 2]}
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_if_empty_template_parameter_is_provided(
  template_query_with_for_loop,
):
  expected_query = 'SELECT field_one, FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_for_loop
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_if_single_template_parameter_is_provided_returns_if_block(
  template_query_with_if_else,
):
  expected_query = 'SELECT field_one, field_two FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_if_else, {'key': 'field_2'}
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_if_single_template_parameter_is_provided_returns_else_block(
  template_query_with_if_else,
):
  expected_query = 'SELECT field_one, field_three FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_if_else, {'key': 'field_3'}
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_if_no_single_template_parameter_is_provided_returns_else_block(
  template_query_with_if_else,
):
  expected_query = 'SELECT field_one, field_three FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_if_else
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_if_empty_single_template_parameter_is_provided_returns_else_block(
  template_query_with_if_else,
):
  expected_query = 'SELECT field_one, field_three FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_if_else, {'key': None}
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_expand_jinja_not_template_parameter_is_provided(
  template_query_with_for_loop,
):
  expected_query = 'SELECT field_one, FROM some_table'
  rendered_query = PostProcessorMixin().expand_jinja(
    template_query_with_for_loop
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_replace_params_template(macro_template_query_with_for_loop):
  expected_query = 'SELECT field_one, field_two AS field_two, 1 AS day_1, 2 AS day_2, FROM some_table'
  rendered_query = PostProcessorMixin().replace_params_template(
    macro_template_query_with_for_loop,
    params={
      'macro': {'field_2': 'field_two'},
      'template': {'cohort_days': '1,2'},
    },
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_replace_params_empty_template_is_ignored(
  macro_template_query_with_for_loop,
):
  expected_query = 'SELECT field_one, field_two AS field_two, FROM some_table'
  rendered_query = PostProcessorMixin().replace_params_template(
    macro_template_query_with_for_loop,
    params={'macro': {'field_2': 'field_two'}},
  )
  assert rendered_query.replace('  ', ' ') == expected_query


def test_replace_empty_params(macro_template_query_with_for_loop):
  expected_query = 'SELECT field_one, {field_2} AS field_two, FROM some_table'
  rendered_query = PostProcessorMixin().replace_params_template(
    macro_template_query_with_for_loop, params={}
  )
  assert rendered_query.replace('  ', ' ') == expected_query



from __future__ import annotations

import dataclasses

from gaarf import base_query


class FakeQueryWithoutInit(base_query.BaseQuery):
  query_text = 'SELECT campaign.id FROM campaign'


class FakeQueryWithInit(base_query.BaseQuery):
  query_text = """
        SELECT campaign.id FROM campaign WHERE campaign.id = {campaign_id}
        """

  def __init__(self, campaign_id: int) -> None:
    self.campaign_id = campaign_id


@dataclasses.dataclass
class FakeQueryDataclass(base_query.BaseQuery):
  query_text = """
        SELECT campaign.id FROM campaign WHERE campaign.id = {campaign_id}
        """
  campaign_id: int


class FakeQueryOldStyle(base_query.BaseQuery):
  def __init__(self, campaign_id: int) -> None:
    self.query_text = f"""
            SELECT campaign.id FROM campaign WHERE campaign.id = {campaign_id}
        """


class TestBaseQuery:
  def test_base_query_correct_init(self):
    expected_query_text = 'SELECT campaign.id FROM campaign'
    fake_query = FakeQueryWithoutInit()
    assert str(fake_query) == expected_query_text.strip()

  def test_base_query_correct_init_with_argument(self):
    expected_query_text = """
        SELECT campaign.id FROM campaign WHERE campaign.id = 1
        """
    fake_query = FakeQueryWithInit(campaign_id=1)
    assert str(fake_query) == expected_query_text.strip()

  def test_base_query_correct_init_with_argument_old_style(self):
    expected_query_text = """
        SELECT campaign.id FROM campaign WHERE campaign.id = 1
        """
    fake_query = FakeQueryOldStyle(campaign_id=1)
    assert str(fake_query) == expected_query_text.strip()

  def test_base_query_correct_init_with_argument_dataclass(self):
    expected_query_text = """
        SELECT campaign.id FROM campaign WHERE campaign.id = 1
        """
    fake_query = FakeQueryDataclass(campaign_id=1)
    assert str(fake_query) == expected_query_text.strip()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from datetime import datetime
from unittest import mock

import pytest
import yaml
from dateutil.relativedelta import relativedelta

from gaarf import api_clients
from gaarf.cli import utils
from gaarf.query_editor import CommonParametersMixin


class TestGaarfConfig:
  @pytest.mark.parametrize(
    'input_account,output_account',
    [
      (
        None,
        None,
      ),
      (
        '123-456-7890',
        '1234567890',
      ),
      (
        [
          '123-456-7890',
        ],
        [
          '1234567890',
        ],
      ),
    ],
  )
  def test_post_init_returns_correctly_formatted_account(
    self, input_account, output_account
  ):
    config = utils.GaarfConfig(
      output='console',
      api_version=api_clients.GOOGLE_ADS_API_VERSION,
      account=input_account,
    )
    assert config.account == output_account

  def test_post_init_returns_correctly_formatted_writer_params(self):
    config = utils.GaarfConfig(
      output='console',
      api_version=api_clients.GOOGLE_ADS_API_VERSION,
      account=None,
      writer_params={
        'page-size': 10,
      },
    )
    assert 'page_size' in config.writer_params


def test_convert_date():
  current_date = datetime.today()
  current_year = datetime(current_date.year, 1, 1)
  current_month = datetime(current_date.year, current_date.month, 1)
  last_year = current_year - relativedelta(years=1)
  last_month = current_month - relativedelta(months=1)
  yesterday = current_date - relativedelta(days=1)

  non_macro_date = '2022-01-01'
  date_year = ':YYYY'
  date_month = ':YYYYMM'
  date_day = ':YYYYMMDD'
  date_year_minus_one = ':YYYY-1'
  date_month_minus_one = ':YYYYMM-1'
  date_day_minus_one = ':YYYYMMDD-1'

  non_macro_date_converted = utils.convert_date(non_macro_date)
  new_date_year = utils.convert_date(date_year)
  new_date_month = utils.convert_date(date_month)
  new_date_day = utils.convert_date(date_day)
  new_date_year_minus_one = utils.convert_date(date_year_minus_one)
  new_date_month_minus_one = utils.convert_date(date_month_minus_one)
  new_date_day_minus_one = utils.convert_date(date_day_minus_one)

  assert non_macro_date_converted == non_macro_date
  assert new_date_year == current_year.strftime('%Y-%m-%d')
  assert new_date_month == current_month.strftime('%Y-%m-%d')
  assert new_date_day == current_date.strftime('%Y-%m-%d')
  assert new_date_year_minus_one == last_year.strftime('%Y-%m-%d')
  assert new_date_month_minus_one == last_month.strftime('%Y-%m-%d')
  assert new_date_day_minus_one == yesterday.strftime('%Y-%m-%d')


def test_wrong_convert_date():
  date_day = ':YYYYMMDD-N'
  with pytest.raises(ValueError):
    utils.convert_date(date_day)


@pytest.fixture
def param_parser():
  return utils.ParamsParser(['macro', 'template'])


def test_if_incorrect_param_is_provided(param_parser):
  with pytest.raises(utils.GaarfParamsException):
    parsed_params = param_parser.parse(
      ['--macros.start_date=2022-01-01', '--fake_param.end_date=2022-12-31']
    )


def test_identify_param_pair_existing(param_parser):
  param_pair = param_parser._identify_param_pair(
    'macro', ['--macro.start_date', '2022-01-01']
  )
  assert param_pair == {'start_date': '2022-01-01'}


def test_identify_param_pair_empty(param_parser):
  param_pair = param_parser._identify_param_pair(
    'macro', ['--missing_param.start_date', '2022-01-01']
  )
  assert param_pair is None


def test_identify_param_pair_raises_error(param_parser):
  with pytest.raises(utils.GaarfParamsException):
    param_parser._identify_param_pair(
      'macro', ['--macro.start_date', ':YYYYMMDD', 'extra_element']
    )


def test_parse_params(param_parser):
  parsed_params = param_parser._parse_params(
    'macro', ['--macro.start_date=2022-01-01', '--macro.end_date=2022-12-31']
  )
  assert parsed_params == {'start_date': '2022-01-01', 'end_date': '2022-12-31'}


def test_parse(param_parser):
  parsed_params = param_parser.parse(
    ['--macro.start_date=2022-01-01', '--macro.end_date=2022-12-31']
  )
  assert parsed_params == {
    'macro': {'start_date': '2022-01-01', 'end_date': '2022-12-31'},
    'template': {},
  }


@pytest.fixture
def config_args():
  from dataclasses import dataclass

  @dataclass
  class FakeConfig:
    customer_id: str
    save: str
    api_version: str
    project: str

  return FakeConfig('1', 'console', '10', 'fake-project')


@pytest.fixture
def config_saver(config_args):
  return utils.ConfigSaver('/tmp/config.yaml')


def test_gaarf_config_saver_gaarf_transforms_comma_separated_account_into_list(
  config_saver,
):
  gaarf_config = utils.GaarfConfig(
    output='console', api_version='10', account='1-,2,3'
  )

  config = config_saver.prepare_config({}, gaarf_config)
  assert config == {
    'gaarf': {
      'account': ['1', '2', '3'],
      'output': 'console',
      'api_version': '10',
    }
  }


def test_gaarf_config_saver_gaarf_dont_save_empty_values(config_saver):
  gaarf_config = utils.GaarfConfig(
    output='console',
    api_version='10',
    account='1',
    params={},
    writer_params={},
    customer_ids_query='',
    customer_ids_query_file='',
  )

  config = config_saver.prepare_config({}, gaarf_config)
  assert config == {
    'gaarf': {
      'account': ['1'],
      'output': 'console',
      'api_version': '10',
    }
  }


def test_gaarf_config_saver_gaarf_dont_save_inner_empty_values(config_saver):
  gaarf_config = utils.GaarfConfig(
    output='console',
    api_version='10',
    account='1',
    params={'macro': {'start_date': ':YYYYMMDD'}},
    writer_params={},
    customer_ids_query='',
    customer_ids_query_file='',
  )

  config = config_saver.prepare_config({}, gaarf_config)
  assert config == {
    'gaarf': {
      'account': ['1'],
      'output': 'console',
      'api_version': '10',
      'params': {'macro': {'start_date': ':YYYYMMDD'}},
    }
  }


def test_config_saver_gaarf_save_customer_ids_query_values(config_saver):
  gaarf_config = utils.GaarfConfig(
    output='console',
    api_version='10',
    account='1',
    params={},
    writer_params={},
    customer_ids_query='SELECT',
    customer_ids_query_file='path/to/file.sql',
  )

  config = config_saver.prepare_config({}, gaarf_config)
  assert config == {
    'gaarf': {
      'account': ['1'],
      'output': 'console',
      'api_version': '10',
      'customer_ids_query': 'SELECT',
      'customer_ids_query_file': 'path/to/file.sql',
    }
  }


def test_config_saver_gaarf_bq(config_saver):
  gaarf_bq_config = utils.GaarfBqConfig(
    project='fake-project',
    dataset_location='fake-location',
    params={'macro': {'bq_project': 'another-fake-project'}},
  )
  config = config_saver.prepare_config({}, gaarf_bq_config)
  assert config == {
    'gaarf-bq': {
      'project': 'fake-project',
      'dataset_location': 'fake-location',
      'params': {'macro': {'bq_project': 'another-fake-project'}},
    }
  }


def test_config_saver_does_not_save_empty_params(config_saver):
  gaarf_bq_config = utils.GaarfBqConfig(
    project='fake-project', dataset_location='fake-location', params={}
  )
  config = config_saver.prepare_config({}, gaarf_bq_config)
  assert config == {
    'gaarf-bq': {'project': 'fake-project', 'dataset_location': 'fake-location'}
  }


def test_config_saver_does_not_save_empty_nested_params(config_saver):
  gaarf_bq_config = utils.GaarfBqConfig(
    project='fake-project',
    dataset_location='fake-location',
    params={
      'macro': {'bq_project': 'another-fake-project'},
    },
  )
  config = config_saver.prepare_config({}, gaarf_bq_config)
  assert config == {
    'gaarf-bq': {
      'project': 'fake-project',
      'dataset_location': 'fake-location',
      'params': {'macro': {'bq_project': 'another-fake-project'}},
    }
  }


def test_initialize_runtime_paramers_injects_common_parameters_to_present_macros(  # pylint: disable=line-to-long
):
  config = utils.GaarfConfig(
    output='console',
    api_version='10',
    account='1',
    params={'macro': {'start_date': ':YYYYMMDD'}},
    writer_params={},
    customer_ids_query=None,
    customer_ids_query_file=None,
  )
  with mock.patch.object(
    CommonParametersMixin,
    'common_params',
    {
      'date_iso': '19700101',
      'current_date': '1970-01-01',
      'current_datetime': '1970-01-01 00-00-00',
    },
  ):
    initialized_config = utils.initialize_runtime_parameters(config)
    expected_config = utils.GaarfConfig(
      output='console',
      api_version='10',
      account='1',
      params={
        'macro': {
          'start_date': datetime.today().strftime('%Y-%m-%d'),
          'date_iso': '19700101',
          'current_date': '1970-01-01',
          'current_datetime': '1970-01-01 00-00-00',
        }
      },
      writer_params={},
      customer_ids_query=None,
      customer_ids_query_file=None,
    )
    assert initialized_config == expected_config


def test_initialize_runtime_paramers_injects_common_parameters_to_empty_macros():  # pylint: disable=line-too-long
  config = utils.GaarfConfig(
    output='console',
    api_version='10',
    account='1',
    params={'macro': {'start_date': '2022-01-01'}},
    writer_params={},
    customer_ids_query=None,
    customer_ids_query_file=None,
  )
  with mock.patch.object(
    CommonParametersMixin,
    'common_params',
    {
      'date_iso': '19700101',
      'current_date': '1970-01-01',
      'current_datetime': '1970-01-01 00-00-00',
    },
  ):
    initialized_config = utils.initialize_runtime_parameters(config)
    expected_config = utils.GaarfConfig(
      output='console',
      api_version='10',
      account='1',
      params={
        'macro': {
          'start_date': '2022-01-01',
          'date_iso': '19700101',
          'current_date': '1970-01-01',
          'current_datetime': '1970-01-01 00-00-00',
        }
      },
      writer_params={},
      customer_ids_query=None,
      customer_ids_query_file=None,
    )
    assert initialized_config == expected_config


def test_initialize_runtime_parameters_overwrites_common_params():
  config = utils.GaarfConfig(
    output='console',
    api_version='10',
    account='1',
    params={
      'macro': {
        'start_date': '2022-01-01',
        'date_iso': '20220101',
      }
    },
    writer_params={},
    customer_ids_query=None,
    customer_ids_query_file=None,
  )
  with mock.patch.object(
    CommonParametersMixin,
    'common_params',
    {
      'date_iso': '19700101',
      'current_date': '1970-01-01',
      'current_datetime': '1970-01-01 00-00-00',
    },
  ):
    initialized_config = utils.initialize_runtime_parameters(config)
    expected_config = utils.GaarfConfig(
      output='console',
      api_version='10',
      account='1',
      params={
        'macro': {
          'start_date': '2022-01-01',
          'date_iso': '20220101',
          'current_date': '1970-01-01',
          'current_datetime': '1970-01-01 00-00-00',
        }
      },
      writer_params={},
      customer_ids_query=None,
      customer_ids_query_file=None,
    )
    assert initialized_config == expected_config


class TestConfigBuilder:
  @pytest.fixture
  def fake_config_path(self, tmp_path):
    config = {
      'gaarf': {
        'output': 'csv',
        'account': '123456789',
        'api_version': 'v16',
        'csv': {
          'destination_folder': '/home',
        },
      },
      'gaarf-bq': {
        'project': 'fake-project',
      },
      'gaarf-sql': {
        'connection_string': 'fake-connection-string',
      },
    }
    file = tmp_path / 'config.yaml'
    with open(file, mode='w', encoding='utf-8') as f:
      yaml.dump(
        config, f, default_flow_style=False, sort_keys=False, encoding='utf-8'
      )
    return file

  def test_build_raises_gaarf_config_exception_when_gaarf_section_is_missing(
    self,
  ):
    with pytest.raises(utils.GaarfConfigException):
      utils.ConfigBuilder(config_type='unknown-type')

  class TestGaarfConfig:
    def test_build_create_valid_config_from_config_file(self, fake_config_path):
      expected_config = utils.GaarfConfig(
        output='csv',
        api_version='v16',
        account='123456789',
        writer_params={'destination_folder': '/home'},
      )

      builder = utils.ConfigBuilder(config_type='gaarf')
      args = {'gaarf_config': fake_config_path}
      kwargs = []
      config = builder.build(args, kwargs)
      assert config == expected_config

    def test_build_create_valid_config_from_cli(self):
      expected_config = utils.GaarfConfig(
        output='csv',
        api_version=api_clients.GOOGLE_ADS_API_VERSION,
        account='123456789',
        params={
          'macro': {
            'start_date': '1970-01-01',
          },
          'template': {
            'skan4': 'true',
          },
        },
        writer_params={'destination_folder': '/srv'},
      )

      builder = utils.ConfigBuilder(config_type='gaarf')
      args = {'output': 'csv', 'account': 123456789}
      kwargs = [
        '--csv.destination-folder=/srv',
        '--macro.start_date=1970-01-01',
        '--template.skan4=true',
      ]
      config = builder.build(args, kwargs)
      assert config == expected_config

    def test_build_create_valid_config_from_config_file_and_cli(
      self, fake_config_path
    ):
      expected_config = utils.GaarfConfig(
        output='csv',
        api_version=api_clients.GOOGLE_ADS_API_VERSION,
        account='123456789',
        params={
          'macro': {
            'start_date': '1970-01-01',
          },
          'template': {
            'skan4': 'true',
          },
        },
        writer_params={'destination_folder': '/srv'},
      )

      builder = utils.ConfigBuilder(config_type='gaarf')
      args = {'gaarf_config': fake_config_path, 'output': 'csv'}
      kwargs = [
        '--csv.destination-folder=/srv',
        '--macro.start_date=1970-01-01',
        '--template.skan4=true',
      ]
      config = builder.build(args, kwargs)
      assert config == expected_config

  class TestGaarfBqConfig:
    def test_build_create_valid_config_file(self, fake_config_path):
      expected_config = utils.GaarfBqConfig(project='fake-project')
      builder = utils.ConfigBuilder(config_type='gaarf-bq')
      args = {'gaarf_config': fake_config_path}
      kwargs = []
      config = builder.build(args, kwargs)
      assert config == expected_config

    def test_build_create_valid_config_from_cli(self):
      expected_config = utils.GaarfBqConfig(
        project='fake-project',
        params={
          'macro': {
            'start_date': '1970-01-01',
          },
          'template': {
            'skan4': 'true',
          },
        },
      )

      builder = utils.ConfigBuilder(config_type='gaarf-bq')
      args = {'project': 'fake-project'}
      kwargs = [
        '--macro.start_date=1970-01-01',
        '--template.skan4=true',
      ]
      config = builder.build(args, kwargs)
      assert config == expected_config

    def test_build_create_valid_config_from_config_file_and_cli(
      self, fake_config_path
    ):
      expected_config = utils.GaarfBqConfig(
        project='new-fake-project',
        params={
          'macro': {
            'start_date': '1970-01-01',
          },
          'template': {
            'skan4': 'true',
          },
        },
      )

      builder = utils.ConfigBuilder(config_type='gaarf-bq')
      args = {'gaarf_config': fake_config_path, 'project': 'new-fake-project'}
      kwargs = [
        '--macro.start_date=1970-01-01',
        '--template.skan4=true',
      ]
      config = builder.build(args, kwargs)
      assert config == expected_config

  class TestGaarfSqlConfig:
    def test_build_create_valid_config_from_config_file(self, fake_config_path):
      expected_config = utils.GaarfSqlConfig(
        connection_string='fake-connection-string'
      )
      builder = utils.ConfigBuilder(config_type='gaarf-sql')
      args = {'gaarf_config': fake_config_path}
      kwargs = []
      config = builder.build(args, kwargs)
      assert config == expected_config

    def test_build_create_valid_config_from_cli(self):
      expected_config = utils.GaarfSqlConfig(
        connection_string='fake-connection-string',
        params={
          'macro': {
            'start_date': '1970-01-01',
          },
          'template': {
            'skan4': 'true',
          },
        },
      )
      builder = utils.ConfigBuilder(config_type='gaarf-sql')
      args = {'connection_string': 'fake-connection-string'}
      kwargs = [
        '--macro.start_date=1970-01-01',
        '--template.skan4=true',
      ]
      config = builder.build(args, kwargs)
      assert config == expected_config

    def test_build_create_valid_config_from_config_file_and_cli(
      self, fake_config_path
    ):
      expected_config = utils.GaarfSqlConfig(
        connection_string='new-fake-connection-string',
        params={
          'macro': {
            'start_date': '1970-01-01',
          },
          'template': {
            'skan4': 'true',
          },
        },
      )
      builder = utils.ConfigBuilder(config_type='gaarf-sql')
      args = {
        'gaarf_config': fake_config_path,
        'connection_string': 'new-fake-connection-string',
      }
      kwargs = [
        '--macro.start_date=1970-01-01',
        '--template.skan4=true',
      ]
      config = builder.build(args, kwargs)
      assert config == expected_config



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

from gaarf import bq_executor


class TestBigQueryExecutor:
  def test_instantiating_bq_executor_is_deprecated(self):
    with pytest.warns(DeprecationWarning) as w:
      bq_executor.BigQueryExecutor(project_id='fake-project')
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)
      assert str(w[0].message) == (
        'Loading BigQueryExecutor from `gaarf.bq_executor` is '
        'deprecated; Import BigQueryExecutor from '
        '`gaarf.executors.bq_executor` instead'
      )


def test_extract_datasets_is_deprecated():
  with pytest.warns(DeprecationWarning) as w:
    bq_executor.extract_datasets({})
    assert len(w) == 1
    assert issubclass(w[0].category, DeprecationWarning)
    assert str(w[0].message) == (
      'Loading `extract_datasets` from `gaarf.bq_executor` is '
      'deprecated; Import `extract_datasets` from '
      '`gaarf.executors.bq_executor` instead'
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging

import pytest
import tenacity
from google.ads.googleads import client as googleads_client
from google.api_core import exceptions as google_exceptions

from gaarf import api_clients


class TestBaseClient:
  @pytest.fixture
  def client(self):
    return api_clients.BaseClient()

  @pytest.mark.parametrize(
    'field_name, possible_values',
    [
      (
        'campaign.name',
        {
          '',
        },
      ),
      (
        'campaign.id',
        {
          0,
        },
      ),
      (
        'metrics.clicks',
        {
          0,
        },
      ),
      (
        'metrics.conversions',
        {
          0.0,
        },
      ),
      (
        'ad_group_ad.ad.video_responsive_ad.videos.asset',
        {
          '',
        },
      ),
      (
        'ad_group_ad.policy_summary.policy_topic_entries',
        {
          '',
        },
      ),
    ],
  )
  def test_infer_types_returns_correct_field_possible_values_for_primitives(
    self, client, field_name, possible_values
  ):
    field_possible_values = client.infer_types(
      field_names=[
        field_name,
      ]
    )
    expected_value = [
      api_clients.FieldPossibleValues(name=field_name, values=possible_values),
    ]
    assert field_possible_values == expected_value

  @pytest.mark.parametrize(
    'field_name, possible_values',
    [
      (
        'segments.device',
        {
          'CONNECTED_TV',
          'DESKTOP',
          'MOBILE',
          'OTHER',
          'TABLET',
          'UNKNOWN',
          'UNSPECIFIED',
        },
      ),
      (
        'ad_group_ad.policy_summary.approval_status',
        {
          'APPROVED',
          'APPROVED_LIMITED',
          'AREA_OF_INTEREST_ONLY',
          'DISAPPROVED',
          'UNKNOWN',
          'UNSPECIFIED',
        },
      ),
    ],
  )
  def test_infer_types_returns_correct_field_possible_values_for_enums(
    self, client, field_name, possible_values
  ):
    field_possible_values = client.infer_types(
      field_names=[
        field_name,
      ]
    )
    expected_value = [
      api_clients.FieldPossibleValues(name=field_name, values=possible_values),
    ]
    assert field_possible_values == expected_value

  def test_get_target_resource_for_metrics_returns_metrics(self, client):
    target_resource = client._get_target_resource('metrics')
    assert target_resource == client._metrics.Metrics

  def test_get_target_resource_for_segment_without_subresource_returns_segments(
    self, client
  ):
    target_resource = client._get_target_resource('segments')
    assert target_resource == client._segments.Segments

  def test_get_target_resource_for_segment_with_subresource_returns_segments(
    self, client
  ):
    sub_resource = 'asset_interaction_target'
    target_resource = client._get_target_resource(
      'segments',
      [
        sub_resource,
      ],
    )
    expected_resource = getattr(
      client._segments, api_clients.clean_resource(sub_resource)
    )
    assert target_resource == expected_resource

  @pytest.mark.skip('Implement')
  def test_get_target_resource_for_segment_with_nested_subresource_returns_segments(
    self, client
  ):
    sub_resource = 'discovery_carousel_card_asset'
    target_resource = client._get_target_resource(
      'asset',
      [
        sub_resource,
      ],
    )
    expected_resource = getattr(
      client._segments, api_clients.clean_resource(sub_resource)
    )
    assert target_resource == expected_resource

  def test_get_decriptor_returns_correct_attribute_name(self, client):
    field = 'segments.date'
    descriptor = client._get_descriptor(field)
    assert descriptor.name == 'date'

  @pytest.mark.parametrize(
    'field_name, value',
    [
      (
        'campaign.name',
        '',
      ),
      (
        'metrics.clicks',
        0,
      ),
      (
        'metrics.conversions_value',
        0.0,
      ),
      (
        'ad_group_ad.ad.added_by_google_ads',
        False,
      ),
    ],
  )
  def test_get_possible_values_for_resource_returns_default_value_for_primitives(
    self, client, field_name, value
  ):
    descriptor = client._get_descriptor(field_name)
    default_value = client._get_possible_values_for_resource(descriptor)
    expected_value = {
      value,
    }
    assert default_value == expected_value

  def test_get_possible_values_for_resource_returns_default_value_for_enums(
    self, client
  ):
    descriptor = client._get_descriptor('segments.device')
    default_value = client._get_possible_values_for_resource(descriptor)
    expected_values = {
      'CONNECTED_TV',
      'DESKTOP',
      'MOBILE',
      'OTHER',
      'TABLET',
      'UNKNOWN',
      'UNSPECIFIED',
    }
    assert default_value == expected_values


class TestGoogleAdsApiClient:
  @pytest.fixture
  def fake_googleads_client(self, mocker):
    mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
    return googleads_client.GoogleAdsClient(
      credentials=None,
      developer_token=None,
      version='v17',
      use_proto_plus=True,
    )

  @pytest.fixture
  def test_client(self, mocker, config_path):
    mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
    mocker.patch(
      f'google.ads.googleads.{api_clients.GOOGLE_ADS_API_VERSION}'
      '.services.services.google_ads_service.GoogleAdsServiceClient'
      '.search_stream',
      side_effect=[
        google_exceptions.InternalServerError('test'),
        google_exceptions.InternalServerError('test'),
        google_exceptions.InternalServerError('test'),
      ],
    )
    return api_clients.GoogleAdsApiClient(path_to_config=config_path)

  def test_get_response_raises_internal_service_error_after_3_failed_retries(
    self, test_client
  ):
    test_client.get_response.retry.wait = tenacity.wait_none()
    with pytest.raises(google_exceptions.InternalServerError):
      test_client.get_response(
        entity_id='1',
        query_text='SELECT customer.id FROM customer',
      )

  def test_from_googleads_client_returns_new_instance(
    self, fake_googleads_client, caplog
  ):
    caplog.set_level(logging.WARNING)

    client = api_clients.GoogleAdsApiClient.from_googleads_client(
      fake_googleads_client
    )

    assert isinstance(client.client, googleads_client.GoogleAdsClient)
    assert client.client.use_proto_plus
    assert client.api_version == fake_googleads_client.version
    assert (
      'Mismatch between values of "use_proto_plus" in '
      'GoogleAdsClient and GoogleAdsApiClient, setting '
      '"use_proto_plus=False"'
    ) not in caplog.text

  def test_from_googleads_client_generates_warning_for_use_proto_plus(
    self, fake_googleads_client, caplog
  ):
    caplog.set_level(logging.WARNING)

    client = api_clients.GoogleAdsApiClient.from_googleads_client(
      fake_googleads_client, use_proto_plus=False
    )

    assert not client.client.use_proto_plus
    assert (
      'Mismatch between values of "use_proto_plus" in '
      'GoogleAdsClient and GoogleAdsApiClient, setting '
      '"use_proto_plus=False"'
    ) in caplog.text



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import itertools
import logging

import pytest
import tenacity
from google.ads.googleads import errors as googleads_exceptions
from google.api_core import exceptions as google_exceptions

from gaarf import (
  api_clients,
  exceptions,
  parsers,
  query_editor,
  report,
  report_fetcher,
)
from tests.unit import helpers

_QUERY = 'SELECT customer.id AS customer_id FROM customer'
_EXPECTED_RESULTS = [
  [1],
  [2],
  [3],
]


class TestAdsReportFetcher:
  @pytest.fixture
  def fake_response(self):
    fake_results = [
      [
        helpers.FakeGoogleAdsRowElement(helpers.Customer(1)),
      ],
      [
        helpers.FakeGoogleAdsRowElement(helpers.Customer(2)),
      ],
      [
        helpers.FakeGoogleAdsRowElement(helpers.Customer(3)),
      ],
    ]
    return helpers.FakeResponse(data=fake_results)

  @pytest.fixture
  def test_client(self, mocker, config_path):
    mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
    return api_clients.GoogleAdsApiClient(path_to_config=config_path)

  @pytest.fixture
  def failing_api_client(self, mocker, config_path):
    mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
    mocker.patch(
      f'google.ads.googleads.{api_clients.GOOGLE_ADS_API_VERSION}'
      '.services.services.google_ads_service.GoogleAdsServiceClient'
      '.search_stream',
      side_effect=[
        google_exceptions.InternalServerError('test'),
        google_exceptions.InternalServerError('test'),
        google_exceptions.InternalServerError('test'),
      ],
    )
    api_client = api_clients.GoogleAdsApiClient(path_to_config=config_path)
    api_client.get_response.retry.wait = tenacity.wait_none()
    return api_client

  @pytest.fixture
  def fake_report_fetcher(self, mocker, test_client, fake_response):
    mocker.patch(
      'gaarf.api_clients.GoogleAdsApiClient.get_response',
      return_value=fake_response,
    )
    return report_fetcher.AdsReportFetcher(test_client)

  def test_parse_ads_response_returns_success(
    self, fake_report_fetcher, caplog
  ):
    caplog.set_level(logging.DEBUG)
    query_specification = query_editor.QuerySpecification(_QUERY).generate()
    parser = parsers.GoogleAdsRowParser(query_specification)
    customer_id = 1
    results = fake_report_fetcher._parse_ads_response(
      query_specification=query_specification,
      parser=parser,
      customer_id=customer_id,
    )

    assert results == _EXPECTED_RESULTS
    assert 'Getting response for query None for customer_id 1' in (caplog.text)

  @pytest.mark.parametrize(
    'strategy',
    [
      'BATCH',
      'BATCH_PROTOBUF',
    ],
  )
  def test_parse_ads_response_in_batches_generates_warning_messages(
    self, fake_report_fetcher, strategy, caplog
  ):
    query_specification = query_editor.QuerySpecification(_QUERY).generate()
    parser = parsers.GoogleAdsRowParser(query_specification)
    customer_id = 1
    optimize_strategy = report_fetcher.OptimizeStrategy[strategy]
    fake_report_fetcher._parse_ads_response(
      query_specification=query_specification,
      parser=parser,
      customer_id=customer_id,
      optimize_strategy=optimize_strategy,
    )

    assert 'Running gaarf in an optimized mode' in caplog.text
    assert f'Optimize strategy is {optimize_strategy.name}' in caplog.text

  def test_fetch_returns_correct_report(self, fake_report_fetcher):
    fetched_report = fake_report_fetcher.fetch(
      query_specification=_QUERY, customer_ids=[1]
    )
    expected_report = report.GaarfReport(
      results=_EXPECTED_RESULTS, column_names=['customer_id']
    )

    assert fetched_report == expected_report

  def test_fetch_raises_gaarf_exception(self, mocker, fake_report_fetcher):
    mocker.patch(
      'gaarf.report_fetcher.AdsReportFetcher._parse_ads_response',
      side_effect=[
        googleads_exceptions.GoogleAdsException(
          'test-error', 'test-call', 'test-failure', 'test-request-id'
        )
      ],
    )
    with pytest.raises(exceptions.GaarfExecutorException, match='test-error'):
      fake_report_fetcher.fetch(query_specification=_QUERY, customer_ids=[1])

  def test_parse_ads_response_sequentially_returns_success(
    self, fake_report_fetcher, fake_response
  ):
    query_specification = query_editor.QuerySpecification(_QUERY).generate()
    parser = parsers.GoogleAdsRowParser(query_specification)
    customer_id = 1
    results = fake_report_fetcher._parse_ads_response_sequentially(
      response=fake_response,
      query_specification=query_specification,
      parser=parser,
      customer_id=customer_id,
    )

    assert results == _EXPECTED_RESULTS

  def test_parse_ads_response_in_batches_returns_success(
    self, fake_report_fetcher, fake_response
  ):
    query_specification = query_editor.QuerySpecification(_QUERY).generate()
    parser = parsers.GoogleAdsRowParser(query_specification)
    customer_id = 1
    results = fake_report_fetcher._parse_ads_response_in_batches(
      response=fake_response,
      query_specification=query_specification,
      parser=parser,
      customer_id=customer_id,
    )

    assert_sequence_content_the_same(results, _EXPECTED_RESULTS)

  def test_parse_ads_response_raises_internal_server_error_after_3_failed_attemps(  # noqa: E501
    self, failing_api_client, caplog
  ):
    fetcher = report_fetcher.AdsReportFetcher(failing_api_client)
    query_specification = query_editor.QuerySpecification(_QUERY).generate()
    parser = parsers.GoogleAdsRowParser(query_specification)
    customer_id = 1
    with pytest.raises(google_exceptions.InternalServerError):
      fetcher._parse_ads_response(
        query_specification=query_specification,
        parser=parser,
        customer_id=customer_id,
      )
    assert (
      'Cannot fetch data from API for query '
      f'"{query_specification.query_title}" 3 times'
    ) in caplog.text


def assert_sequence_content_the_same(
  results: list[list[parsers.GoogleAdsRowElement]],
  other_results: list[list[parsers.GoogleAdsRowElement]],
) -> bool:
  return sorted(itertools.chain.from_iterable(results)) == sorted(
    itertools.chain.from_iterable(other_results)
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for executing queries in BigQuery."""

from __future__ import annotations

try:
  from google.cloud import bigquery  # type: ignore
except ImportError as e:
  raise ImportError(
    'Please install google-ads-api-report-fetcher with BigQuery support '
    '- `pip install google-ads-api-report-fetcher[bq]`'
  ) from e

import logging

import pandas as pd
from google.cloud import exceptions as google_cloud_exceptions

from gaarf import query_post_processor

logger = logging.getLogger(__name__)


class BigQueryExecutorException(Exception):
  """Error when executor fails to run query."""


class BigQueryExecutor(query_post_processor.PostProcessorMixin):
  """Handles query execution in BigQuery.

  Attributes:
      project_id: Google Cloud project id.
      location: BigQuery dataset location.
      client: BigQuery client.
  """

  def __init__(self, project_id: str, location: str | None = None) -> None:
    """Initializes BigQueryExecutor.

    Args:
        project_id: Google Cloud project id.
        location: BigQuery dataset location.
    """
    self.project_id = project_id
    self.location = location

  @property
  def client(self) -> bigquery.Client:
    """Instantiates bigquery client."""
    return bigquery.Client(self.project_id)

  def execute(
    self, script_name: str, query_text: str, params: dict | None = None
  ) -> pd.DataFrame:
    """Executes query in BigQuery.

    Args:
        script_name: Script identifier.
        query_text: Query to be executed.
        params: Optional parameters to be replaced in query text.

    Returns:
        DataFrame if query returns some data otherwise empty DataFrame.
    """
    query_text = self.replace_params_template(query_text, params)
    job = self.client.query(query_text)
    try:
      result = job.result()
      logger.debug('%s launched successfully', script_name)
      if result.total_rows:
        return result.to_dataframe()
      return pd.DataFrame()
    except google_cloud_exceptions.GoogleCloudError as e:
      raise BigQueryExecutorException(e) from e

  def create_datasets(self, macros: dict | None) -> None:
    """Creates datasets in BQ based on values in a dict.

    If dict contains keys with 'dataset' in them, then values for such keys
    are treated as dataset names.

    Args:
        macros: Mapping containing data for query execution.
    """
    if macros and (datasets := extract_datasets(macros)):
      for dataset in datasets:
        dataset_id = f'{self.project_id}.{dataset}'
        try:
          self.client.get_dataset(dataset_id)
        except google_cloud_exceptions.NotFound:
          bq_dataset = bigquery.Dataset(dataset_id)
          bq_dataset.location = self.location
          self.client.create_dataset(bq_dataset, timeout=30)
          logger.debug('Created new dataset %s', dataset_id)


def extract_datasets(macros: dict | None) -> list[str]:
  """Finds dataset-related keys based on values in a dict.

  If dict contains keys with 'dataset' in them, then values for such keys
  are treated as dataset names.

  Args:
      macros: Mapping containing data for query execution.

  Returns:
      Possible names of datasets.
  """
  if not macros:
    return []
  return [value for macro, value in macros.items() if 'dataset' in macro]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest
import sqlalchemy

from gaarf import sql_executor


class TestSqlAlchemyQueryExecutor:
  @pytest.fixture
  def engine(self):
    return sqlalchemy.create_engine('sqlite:///:memory:')

  def test_instantiating_sql_executor_is_deprecated(self, engine):
    with pytest.warns(DeprecationWarning) as w:
      sql_executor.SqlAlchemyQueryExecutor(engine)
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)



from __future__ import annotations

import pytest

from gaarf.io import formatter
from gaarf.report import GaarfReport


class TestFormatterWithoutPlaceholders:
  @pytest.fixture
  def report_without_arrays(self):
    return GaarfReport(
      results=[
        [1, 2],
        [2, 3],
        [3, 4],
      ],
      column_names=['campaign_id', 'ad_group_id'],
    )

  @pytest.fixture
  def report_with_arrays(self):
    return GaarfReport(
      results=[
        [1, [2]],
        [2, [3]],
        [3, [4, 5]],
      ],
      column_names=['campaign_id', 'ad_group_id'],
    )

  def test_format_report_for_writing_keeps_arrays_in_report_(
    self, report_with_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy(type_='arrays')
    formatted_report = formatter.format_report_for_writing(
      report_with_arrays, [array_handling_strategy]
    )
    assert report_with_arrays == formatted_report

  def test_format_report_for_writing_returns_the_same_report_under_default_array_handling_strategy(  # pylint: disable=line-too-long
    self, report_without_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy()
    formatted_report = formatter.format_report_for_writing(
      report_without_arrays, [array_handling_strategy]
    )
    assert report_without_arrays == formatted_report

  def test_format_report_for_writing_converts_arrays_converted_to_strings(
    self, report_with_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy(type_='strings')
    formatted_report = formatter.format_report_for_writing(
      report_with_arrays, [array_handling_strategy]
    )
    expected_report = GaarfReport(
      results=[
        [1, '2'],
        [2, '3'],
        [3, '4|5'],
      ],
      column_names=['campaign_id', 'ad_group_id'],
      results_placeholder=[[0, '']],
    )
    assert expected_report == formatted_report

  def test_format_report_for_writing_converts_arrays_converted_to_strings_with_custom_delimiter(  # pylint: disable=line-too-long
    self, report_with_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy(
      type_='strings', delimiter='*'
    )
    formatted_report = formatter.format_report_for_writing(
      report_with_arrays, [array_handling_strategy]
    )
    expected_report = GaarfReport(
      results=[
        [1, '2'],
        [2, '3'],
        [3, '4*5'],
      ],
      column_names=['campaign_id', 'ad_group_id'],
    )
    assert expected_report == formatted_report


class TestFormatterWithPlaceholders:
  @pytest.fixture
  def report_without_arrays(self):
    return GaarfReport(
      results=[
        [1, 2],
        [2, 3],
        [3, 4],
      ],
      column_names=['campaign_id', 'ad_group_id'],
      results_placeholder=[[0, 0]],
    )

  @pytest.fixture
  def report_with_arrays(self):
    return GaarfReport(
      results=[
        [1, [2]],
        [2, [3]],
        [3, [4, 5]],
      ],
      column_names=['campaign_id', 'ad_group_id'],
      results_placeholder=[[0, [0]]],
    )

  def test_format_report_for_writing_keeps_placeholders_the_same(
    self, report_with_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy(type_='arrays')
    formatted_report = formatter.format_report_for_writing(
      report_with_arrays, [array_handling_strategy]
    )
    assert report_with_arrays.results_placeholder == (
      formatted_report.results_placeholder
    )

  def test_format_report_for_writing_keeps_placeholders_the_same_for_default_array_handling_strategy(  # pylint: disable=line-too-long
    self, report_without_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy()
    formatted_report = formatter.format_report_for_writing(
      report_without_arrays, [array_handling_strategy]
    )
    assert report_without_arrays.results_placeholder == (
      formatted_report.results_placeholder
    )

  def test_format_report_for_writing_convert_array_placeholders_to_strings(
    self, report_with_arrays
  ):
    array_handling_strategy = formatter.ArrayHandlingStrategy(type_='strings')
    formatted_report = formatter.format_report_for_writing(
      report_with_arrays, [array_handling_strategy]
    )
    expected_report = GaarfReport(
      results=[
        [1, '2'],
        [2, '3'],
        [3, '4|5'],
      ],
      column_names=['campaign_id', 'ad_group_id'],
      results_placeholder=[[0, '']],
    )
    assert expected_report.results_placeholder == (
      formatted_report.results_placeholder
    )


def test_format_extension_returns_correct_extensions():
  default_output = formatter.format_extension('test_query.sql')
  default_output_custom_extension = formatter.format_extension(
    'test_query.txt', '.txt'
  )
  csv_output = formatter.format_extension(
    'test_query.sql', new_extension='.csv'
  )
  assert default_output == 'test_query'
  assert default_output_custom_extension == 'test_query'
  assert csv_output == 'test_query.csv'



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data to Google Sheets."""

from __future__ import annotations

from google.auth import exceptions as auth_exceptions

try:
  import gspread
except ImportError as e:
  raise ImportError(
    'Please install google-ads-api-report-fetcher with sheets support - '
    '`pip install google-ads-api-report-fetcher[sheet]`'
  ) from e

import datetime
import logging

from gaarf.io import formatter
from gaarf.io.writers.abs_writer import AbsWriter
from gaarf.report import GaarfReport


class SheetWriter(AbsWriter):
  def __init__(
    self,
    share_with: str,
    credentials_file: str,
    spreadsheet_url: str | None = None,
    is_append: bool = False,
    **kwargs: str,
  ) -> None:
    """Initialize the SheetWriter to write reports to Google Sheets.

    Args:
        share_with: Email address to share the spreadsheet.
        credentials_file: Path to the service account credentials file.
        spreadsheet_url: URL of the Google Sheets spreadsheet.
        is_append: Whether you want to append data to the spreadsheet.
    """
    super().__init__(**kwargs)
    self.share_with = share_with
    self.credentials_file = credentials_file
    self.spreadsheet_url = spreadsheet_url
    self.is_append = is_append
    self.client = None
    self.spreadsheet = None
    self.gspread_client = None

  def write(
    self,
    report: GaarfReport,
    destination: str = f'Report {datetime.datetime.utcnow()}',
  ) -> str:
    self._init_client()
    report = self.format_for_write(report)
    if not destination:
      destination = f'Report {datetime.datetime.utcnow()}'
    destination = formatter.format_extension(destination)
    num_data_rows = len(report) + 1
    try:
      sheet = self.spreadsheet.worksheet(destination)
    except gspread.exceptions.WorksheetNotFound:
      sheet = self.spreadsheet.add_worksheet(
        destination, rows=num_data_rows, cols=len(report.column_names)
      )
    if not self.is_append:
      sheet.clear()
      self._add_rows_if_needed(num_data_rows, sheet)
      sheet.append_rows(
        [report.column_names] + report.results, value_input_option='RAW'
      )
    else:
      self._add_rows_if_needed(num_data_rows, sheet)
      sheet.append_rows(report.results, value_input_option='RAW')

    success_msg = f'Report is saved to {sheet.url}'
    logging.info(success_msg)
    if self.share_with:
      self.spreadsheet.share(self.share_with, perm_type='user', role='writer')
    return success_msg

  def _init_client(self) -> None:
    if not self.client:
      if not self.credentials_file:
        raise ValueError(
          'Provide path to service account via ' '`credentials_file` option'
        )
      try:
        self.gspread_client = gspread.service_account(
          filename=self.credentials_file
        )
      except auth_exceptions.MalformedError:
        self.gspread_client = gspread.oauth(
          credentials_filename=self.credentials_file
        )
      self._open_sheet()

  def _open_sheet(self) -> None:
    if not self.spreadsheet:
      if not self.spreadsheet_url:
        self.spreadsheet = self.gspread_client.create(
          f'Gaarf CSV {datetime.datetime.utcnow()}'
        )
      else:
        self.spreadsheet = self.gspread_client.open_by_url(self.spreadsheet_url)

  def _add_rows_if_needed(
    self, num_data_rows: int, sheet: gspread.worksheet.Worksheet
  ) -> None:
    num_sheet_rows = len(sheet.get_all_values())
    if num_data_rows > num_sheet_rows:
      num_rows_to_add = num_data_rows - num_sheet_rows
      sheet.add_rows(num_rows_to_add)

  def __str__(self) -> str:
    return f'[Sheet] - data are saved to {self.spreadsheet_url}.'



from __future__ import annotations

import pytest

from gaarf import api_clients, builtin_queries, report, report_fetcher


class TestBuiltInQueries:
  @pytest.fixture
  def test_client(self, mocker, config_path):
    mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
    return api_clients.GoogleAdsApiClient(path_to_config=config_path)

  @pytest.fixture
  def fake_report_fetcher(self, mocker, test_client):
    data = [
      [
        1,
        'https://ads.google.com/aw/accounts?__u=1&&ocid=12345',
      ],
    ]
    column_names = [
      'account_id',
      'url',
    ]
    mocker.patch(
      'gaarf.report_fetcher.AdsReportFetcher.fetch',
      return_value=report.GaarfReport(data, column_names),
    )
    return report_fetcher.AdsReportFetcher(test_client)

  @pytest.fixture
  def missing_ocid_fake_report_fetcher(self, mocker, test_client):
    data = [
      [
        1,
        '',
      ],
    ]
    column_names = [
      'account_id',
      'url',
    ]
    mocker.patch(
      'gaarf.report_fetcher.AdsReportFetcher.fetch',
      return_value=report.GaarfReport(data, column_names),
    )
    return report_fetcher.AdsReportFetcher(test_client)

  @pytest.fixture
  def empty_fake_report_fetcher(self, mocker, test_client):
    column_names = [
      'account_id',
      'url',
    ]
    mocker.patch(
      'gaarf.report_fetcher.AdsReportFetcher.fetch',
      return_value=report.GaarfReport([], column_names),
    )
    return report_fetcher.AdsReportFetcher(test_client)

  def test_get_ocid_mapping_returns_correct_result(self, fake_report_fetcher):
    account_id = 1
    expected_report = report.GaarfReport(
      results=[
        [
          account_id,
          '12345',
        ],
      ],
      column_names=[
        'account_id',
        'ocid',
      ],
    )
    ocid_mapping = builtin_queries.get_ocid_mapping(
      fake_report_fetcher,
      accounts=[
        account_id,
      ],
    )
    assert ocid_mapping == expected_report

  def test_get_ocid_mapping_returns_placeholder_for_missing_ocid_report(
    self, missing_ocid_fake_report_fetcher
  ):
    account_id = 1
    expected_report = report.GaarfReport(
      results=[
        [
          account_id,
          '0',
        ],
      ],
      column_names=[
        'account_id',
        'ocid',
      ],
    )
    ocid_mapping = builtin_queries.get_ocid_mapping(
      missing_ocid_fake_report_fetcher,
      accounts=[
        account_id,
      ],
    )
    assert ocid_mapping == expected_report

  def test_get_ocid_mapping_returns_placeholder_for_empty_report(
    self, empty_fake_report_fetcher
  ):
    account_id = 1
    expected_report = report.GaarfReport(
      results=[
        [
          account_id,
          '0',
        ],
      ],
      column_names=[
        'account_id',
        'ocid',
      ],
    )
    ocid_mapping = builtin_queries.get_ocid_mapping(
      empty_fake_report_fetcher,
      accounts=[
        account_id,
      ],
    )
    assert ocid_mapping == expected_report



from __future__ import annotations

from gaarf.executors import bq_executor


def test_extract_datasets():
  macros = {
    'start_date': ':YYYYMMDD',
    'bq_dataset': 'dataset_1',
    'dataset_new': 'dataset_2',
    'legacy_dataset_old': 'dataset_3',
    'wrong_dts': 'dataset_4',
  }

  expected = [
    'dataset_1',
    'dataset_2',
    'dataset_3',
  ]
  datasets = bq_executor.extract_datasets(macros)
  assert datasets == expected



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections import abc

import pandas as pd
import pytest

from gaarf import exceptions, report


@pytest.fixture
def single_element_report():
  return report.GaarfReport(results=[[1]], column_names=['campaign_id'])


@pytest.fixture
def single_column_report():
  return report.GaarfReport(
    results=[[1], [1], [3]], column_names=['campaign_id']
  )


@pytest.fixture
def multi_column_report():
  return report.GaarfReport(
    results=[[1, 2], [2, 3], [3, 4]],
    column_names=['campaign_id', 'ad_group_id'],
  )


class TestGaarfReport:
  class TestGaarfReportIteration:
    def test_single_element_report_returns_gaarf_row(
      self, single_element_report
    ):
      assert [row[0] for row in single_element_report] == [1]

    def test_single_column_report_returns_sequence(self, single_column_report):
      assert [row[0] for row in single_column_report] == [1, 1, 3]

    def test_multi_column_report_returns_gaarf_row(self, multi_column_report):
      assert isinstance(list(multi_column_report)[0], report.GaarfRow)

    def test_multi_column_report_support_iteration_with_gaarf_iterator(
      self,
      multi_column_report,
    ):
      assert isinstance(iter(multi_column_report), abc.Iterable)

    def test_multi_column_report_get_element_by_id(self, multi_column_report):
      assert [row[0] for row in multi_column_report] == [1, 2, 3]

    def test_multi_column_report_get_element_by_name(self, multi_column_report):
      assert [row['campaign_id'] for row in multi_column_report] == [1, 2, 3]

    def test_multi_column_report_get_element_with_get_method(
      self, multi_column_report
    ):
      assert [row.get('campaign_id') for row in multi_column_report] == [
        1,
        2,
        3,
      ]

    def test_multi_column_report_get_element_as_attribute(
      self, multi_column_report
    ):
      assert [row.campaign_id for row in multi_column_report] == [1, 2, 3]

    def test_getitem_raise_index_error_for_out_of_index_value(
      self,
      multi_column_report,
    ):
      with pytest.raises(exceptions.GaarfReportException):
        [row[99] for row in multi_column_report] == [1, 2, 3]

    def test_get_raises_attribute_error_for_missing_value(
      self, multi_column_report
    ):
      with pytest.raises(AttributeError):
        [row.get('missing_value') for row in multi_column_report] == [
          None,
          None,
          None,
        ]

    def test_getattr_raises_attribute_error_for_missing_value(
      self,
      multi_column_report,
    ):
      with pytest.raises(AttributeError):
        [getattr(row, 'missing_value') for row in multi_column_report] == [
          None,
          None,
          None,
        ]

    def test_hasattr_return_false_for_missing_value(self, multi_column_report):
      assert [hasattr(row, 'missing_value') for row in multi_column_report] == [
        False,
        False,
        False,
      ]

  class TestGaarfReportMisc:
    def test_get_report_length(self, multi_column_report):
      assert len(multi_column_report) == len(multi_column_report.results)

    def test_report_bool(self, single_element_report):
      assert single_element_report
      single_element_report.results = []
      assert not single_element_report

  class TestGaarfReportAddition:
    def test_add_two_reports(self, multi_column_report):
      added_report = multi_column_report + multi_column_report
      assert len(added_report) == len(multi_column_report.results) * 2

    def test_add_report_and_non_report_raises_exception(
      self, multi_column_report
    ):
      with pytest.raises(exceptions.GaarfReportException):
        multi_column_report + 1

    def test_add_non_report_and_report_raises_exception(
      self, multi_column_report
    ):
      with pytest.raises(TypeError):
        1 + multi_column_report

    def test_add_reports_with_different_columns_raises_exception(
      self, multi_column_report, single_element_report
    ):
      with pytest.raises(exceptions.GaarfReportException):
        multi_column_report + single_element_report

  class TestGaarfReportSlicing:
    def test_slicing_multi_column_gaarf_report_returns_gaarf_report(
      self,
      multi_column_report,
    ):
      new_report = multi_column_report[0:2]
      assert new_report == report.GaarfReport(
        results=[[1, 2], [2, 3]], column_names=['campaign_id', 'ad_group_id']
      )

    def test_indexing_multi_column_gaarf_report_by_single_index_returns_gaarf_row(
      self,
      multi_column_report,
    ):
      new_report = multi_column_report[0]
      assert new_report == report.GaarfRow(
        data=[1, 2], column_names=['campaign_id', 'ad_group_id']
      )

    def test_indexing_multi_column_gaarf_report_by_multi_index_returns_gaarf_report(
      self,
      multi_column_report,
    ):
      new_report = multi_column_report[0:2]
      assert new_report == report.GaarfReport(
        results=[[1, 2], [2, 3]], column_names=['campaign_id', 'ad_group_id']
      )

    def test_indexing_multi_column_gaarf_report_by_one_column_returns_gaarf_report(
      self,
      multi_column_report,
    ):
      new_report = multi_column_report['campaign_id']
      assert new_report == report.GaarfReport(
        results=[[1], [2], [3]], column_names=['campaign_id']
      )

    def test_indexing_multi_column_gaarf_report_by_several_columns_returns_gaarf_report(
      self,
      multi_column_report,
    ):
      new_report = multi_column_report[['campaign_id', 'ad_group_id']]
      assert new_report == multi_column_report

    def test_indexing_multi_column_gaarf_report_by_non_existing_column_raises_exception(
      self,
      multi_column_report,
    ):
      with pytest.raises(exceptions.GaarfReportException):
        multi_column_report[['campaign_id', 'ad_group']]

    def test_slicing_single_column_gaarf_report_returns_report(
      self,
      single_column_report,
    ):
      single_column_report.disable_scalar_conversions()
      new_report = single_column_report[0:2]
      expected_report = report.GaarfReport(
        results=[[1], [1]], column_names=['campaign_id']
      )
      assert new_report == expected_report

    def test_slicing_single_column_gaarf_report_returns_row(
      self,
      single_column_report,
    ):
      single_column_report.disable_scalar_conversions()
      row = single_column_report[0]
      expected_row = report.GaarfRow(data=[1], column_names=['campaign_id'])
      assert row == expected_row

    def test_slicing_single_column_gaarf_report_returns_slice(
      self,
      single_column_report,
    ):
      with pytest.warns(FutureWarning) as w:
        result = single_column_report[0:2]
        assert result == [1, 1]
        assert len(w) == 1
        assert str(w[0].message) == (
          'Getting scalars from single column `GaarfReport` is discouraged and '
          'will be deprecated in future releases of gaarf. To get scalar value '
          'use `get_value()` method instead. '
          'Call `.disable_scalar_conversions()` to return GaarfRow '
          'or GaarfReport.'
        )

    def test_slicing_single_column_gaarf_report_returns_element(
      self,
      single_column_report,
    ):
      with pytest.warns(FutureWarning) as w:
        result = single_column_report[0]
        assert result == [1]
        assert len(w) == 1
        assert str(w[0].message) == (
          'Getting scalars from single column `GaarfReport` is discouraged and '
          'will be deprecated in future releases of gaarf. To get scalar value '
          'use `get_value()` method instead. '
          'Call `.disable_scalar_conversions()` to return GaarfRow '
          'or GaarfReport.'
        )

    def test_set_existing_attribute_gaarf_multiple_rows_updates_columns(
      self,
      multi_column_report,
    ):
      for row in multi_column_report:
        row.campaign_id = row.campaign_id * 100
      assert multi_column_report == report.GaarfReport(
        results=[[100, 2], [200, 3], [300, 4]],
        column_names=['campaign_id', 'ad_group_id'],
      )

    def test_set_non_existing_attribute_gaarf_rows_get_new_columns(
      self,
      multi_column_report,
    ):
      for row in multi_column_report:
        row.campaign_id_new = row.campaign_id * 100
      assert multi_column_report == report.GaarfReport(
        results=[[1, 2, 100], [2, 3, 200], [3, 4, 300]],
        column_names=['campaign_id', 'ad_group_id', 'campaign_id_new'],
      )

  class TestGaarfReportConversions:
    def test_single_column_report_returns_flattened_list(
      self, single_column_report
    ):
      assert single_column_report.to_list() == [1, 1, 3]

    def test_single_column_report_returns_distinct_flattened_list(
      self,
      single_column_report,
    ):
      assert single_column_report.to_list(row_type='scalar', distinct=True) == [
        1,
        3,
      ]

    def test_single_column_report_returns_distinct_flattened_list_legacy(
      self,
      single_column_report,
    ):
      assert single_column_report.to_list(flatten=True, distinct=True) == [1, 3]

    def test_multi_column_report_converted_to_dict_list_values(
      self,
      multi_column_report,
    ):
      key_column = 'campaign_id'
      value_column = 'ad_group_id'
      output_dict = multi_column_report.to_dict(
        key_column=key_column, value_column=value_column
      )
      assert output_dict == {1: [2], 2: [3], 3: [4]}

    def test_multi_column_report_converted_to_dict_scalar_values(
      self,
      multi_column_report,
    ):
      key_column = 'campaign_id'
      value_column = 'ad_group_id'
      output_dict = multi_column_report.to_dict(
        key_column=key_column,
        value_column=value_column,
        value_column_output='scalar',
      )
      assert output_dict == {1: 2, 2: 3, 3: 4}

    def test_multi_column_report_converted_to_dict_raises_exception_on_non_existing_key_column(
      self,
      multi_column_report,
    ):
      key_column = 'missing_column'
      value_column = 'ad_group_id'
      with pytest.raises(exceptions.GaarfReportException):
        multi_column_report.to_dict(
          key_column=key_column,
          value_column=value_column,
          value_column_output='scalar',
        )

    def test_multi_column_report_converted_to_dict_raises_exception_on_non_existing_value_column(
      self,
      multi_column_report,
    ):
      key_column = 'campaign_id'
      value_column = 'missing_column'
      with pytest.raises(exceptions.GaarfReportException):
        multi_column_report.to_dict(
          key_column=key_column,
          value_column=value_column,
          value_column_output='scalar',
        )

    def test_multi_column_report_converted_to_dict_with_missing_value_column(
      self,
      multi_column_report,
    ):
      key_column = 'campaign_id'
      output_dict = multi_column_report.to_dict(key_column=key_column)
      assert output_dict == {
        1: [{'campaign_id': 1, 'ad_group_id': 2}],
        2: [{'campaign_id': 2, 'ad_group_id': 3}],
        3: [{'campaign_id': 3, 'ad_group_id': 4}],
      }

    def test_multi_column_report_converted_to_dict_without_arguments(
      self,
      multi_column_report,
    ):
      output_dict = multi_column_report.to_list(row_type='dict')
      assert output_dict == [
        {'campaign_id': 1, 'ad_group_id': 2},
        {'campaign_id': 2, 'ad_group_id': 3},
        {'campaign_id': 3, 'ad_group_id': 4},
      ]

    def test_to_list_incorrect_row_type_raises_exception(
      self, multi_column_report
    ):
      with pytest.raises(exceptions.GaarfReportException):
        multi_column_report.to_list(row_type='tuple')

    def test_empty_report_converted_to_dict_with_key_column(
      self, multi_column_report
    ):
      key_column = 'campaign_id'
      value_column = 'ad_group_id'
      # clear results of report
      multi_column_report.results = []
      output_dict = multi_column_report.to_dict(
        key_column=key_column,
        value_column=value_column,
        value_column_output='scalar',
      )
      assert output_dict == {key_column: None}

    def test_conversion_from_pandas(
      self,
    ):
      values = [[1, 2], [3, 4]]
      column_names = ['one', 'two']
      df = pd.DataFrame(data=values, columns=column_names)
      report_from_df = report.GaarfReport.from_pandas(df)
      expected_report = report.GaarfReport(
        results=values, column_names=column_names
      )
      assert report_from_df == expected_report

    def test_convert_report_to_pandas(self, multi_column_report):
      expected = pd.DataFrame(
        data=[[1, 2], [2, 3], [3, 4]], columns=['campaign_id', 'ad_group_id']
      )
      assert multi_column_report.to_pandas().equals(expected)

    def test_get_value_single_element_report_returns_correct_value(
      self,
      single_element_report,
    ):
      assert single_element_report.get_value() == 1

    def test_get_value_multi_column_report_returns_index_position(
      self,
      multi_column_report,
    ):
      assert (
        multi_column_report.get_value(column_index=1, row_index=1)
        == multi_column_report[1][1]
      )

    def test_get_value_raises_exception_when_column_index_out_of_bound(
      self,
      single_element_report,
    ):
      with pytest.raises(exceptions.GaarfReportException):
        single_element_report.get_value(column_index=1)

    def test_get_value_raises_exception_when_row_index_out_of_bound(
      self,
      single_element_report,
    ):
      with pytest.raises(exceptions.GaarfReportException):
        single_element_report.get_value(row_index=1)

  class TestGaarfReportEquality:
    def test_report_with_different_columns_not_equal(
      self, single_element_report, multi_column_report
    ):
      assert single_element_report != multi_column_report

    def test_report_with_different_data_are_not_equal(
      self, multi_column_report
    ):
      new_multi_column_report = report.GaarfReport(
        results=list(multi_column_report.results),
        column_names=list(multi_column_report.column_names),
      )
      new_multi_column_report.results[0] = [10, 10]
      assert new_multi_column_report != multi_column_report

    def test_report_with_same_data_are_equal(self, multi_column_report):
      new_multi_column_report = report.GaarfReport(
        results=list(multi_column_report.results),
        column_names=list(multi_column_report.column_names),
      )
      assert new_multi_column_report == multi_column_report

    def test_report_with_same_data_different_order_are_equal(
      self, multi_column_report
    ):
      new_multi_column_report = report.GaarfReport(
        results=[[2, 1], [3, 2], [4, 3]],
        column_names=['ad_group_id', 'campaign_id'],
      )
      assert new_multi_column_report == multi_column_report


class TestGaarfRow:
  @pytest.fixture
  def test_row(self, multi_column_report):
    return multi_column_report[0]

  def test_set_non_existing_item_gaarf_row_get_new_column(self, test_row):
    test_row['campaign_id_new'] = test_row['campaign_id'] * 100
    assert test_row == report.GaarfRow(
      data=[1, 2, 100],
      column_names=['campaign_id', 'ad_group_id', 'campaign_id_new'],
    )

  def test_set_existing_item_gaarf_row_updates_column(self, test_row):
    test_row['campaign_id'] = test_row['campaign_id'] * 100
    assert test_row == report.GaarfRow(
      data=[100, 2], column_names=['campaign_id', 'ad_group_id']
    )

  def test_set_non_existing_attribute_gaarf_test_row_get_new_column(
    self,
    test_row,
  ):
    test_row.campaign_id_new = test_row.campaign_id * 100
    assert test_row == report.GaarfRow(
      data=[1, 2, 100],
      column_names=['campaign_id', 'ad_group_id', 'campaign_id_new'],
    )

  def test_set_existing_attribute_gaarf_test_row_updates_column(self, test_row):
    test_row.campaign_id = test_row.campaign_id * 100
    assert test_row == report.GaarfRow(
      data=[100, 2], column_names=['campaign_id', 'ad_group_id']
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import json

import proto
import pytest
from google.ads.googleads.v16.resources.types import (
  ad_group_ad_asset_view,
  change_event,
)

from gaarf import exceptions, parsers, query_editor


@dataclasses.dataclass
class FakeGoogleAdsRowElement:
  value: int
  text: str
  name: str


@dataclasses.dataclass
class TextAttribute:
  text: str


@dataclasses.dataclass
class NameAttribute:
  name: str


@dataclasses.dataclass
class AssetAttribute:
  name: str


@dataclasses.dataclass
class ValueAttribute:
  value: str


@dataclasses.dataclass
class Metric:
  clicks: int
  impressions: int


@dataclasses.dataclass(frozen=True)
class NestedResource:
  nested_field: str


@dataclasses.dataclass(frozen=True)
class FakeAdsRowMultipleElements:
  campaign_type: NameAttribute
  clicks: int
  resource: str
  value: str
  metrics: Metric
  old_resource: NestedResource
  new_resource: change_event.ChangeEvent.ChangedResource
  policy_summary: ad_group_ad_asset_view.AdGroupAdAssetPolicySummary

  @property
  def _pb(self):
    return self


@dataclasses.dataclass
class FakeQuerySpecification:
  customizers: dict[str, str]
  virtual_columns: dict[str, str]
  fields: list[str]
  column_names: list[str]


@pytest.fixture
def fake_query_specification():
  customizers = {
    'resource': {
      'type': 'resource_index',
      'value': 0,
    },
    'old_resource': {
      'type': 'nested_field',
      'value': 'nested_field',
    },
    'new_resource': {
      'type': 'nested_field',
      'value': 'campaign.target_cpa.target_cpa_micros',
    },
    'policy_summary': {
      'type': 'nested_field',
      'value': 'policy_topic_entries.type',
    },
    'approval_status': {
      'type': 'nested_field',
      'value': 'approval_status',
    },
  }
  virtual_columns = {
    'date': {
      'type': 'built-in',
      'value': 'date',
    }
  }
  return FakeQuerySpecification(
    customizers=customizers,
    virtual_columns=virtual_columns,
    fields=[
      'campaign_type',
      'clicks',
      'resource',
      'value',
      'old_resource',
      'new_resource',
      'policy_summary',
      'policy_summary',
    ],
    column_names=[
      'campaign_type',
      'clicks',
      'resource',
      'value',
      'old_resource',
      'new_resource',
      'policy_summary',
      'approval_status',
    ],
  )


@pytest.fixture
def fake_google_ads_row_element():
  return FakeGoogleAdsRowElement(1, '2', '3')


@pytest.fixture
def google_ads_row_parser(fake_query_specification):
  return parsers.GoogleAdsRowParser(fake_query_specification)


class FakeMessage(proto.Message):
  message = proto.Field(proto.STRING, number=1)


class TestResourceFormatter:
  def test_get_resource_returns_correct_resource(self):
    resource = parsers.ResourceFormatter('name: id').get_nested_resource()
    assert resource.element == 'id'

  def test_get_resource_id_returns_correct_resource_id(self):
    resource = parsers.ResourceFormatter(
      'customers/1/resource/2'
    ).get_resource_id()
    assert resource.element == '2'

  def test_clear_resource_id_returns_integer_if_possible(self):
    resource = parsers.ResourceFormatter('"1"').clean_resource_id()
    assert resource.element == 1

  def test_clear_resource_id_returns_str_as_default(self):
    resource = parsers.ResourceFormatter('"value"').clean_resource_id()
    assert resource.element == 'value'

  def test_format_returns_correct_result(self):
    resource = (
      parsers.ResourceFormatter('name: id')
      .get_nested_resource()
      .get_resource_id()
      .clean_resource_id()
      .format()
    )
    assert resource == 'id'


class TestParser:
  @pytest.fixture
  def base_parser(self):
    return parsers.BaseParser(None)

  @pytest.fixture
  def attribute_parser(self, base_parser):
    return parsers.AttributeParser(base_parser)

  @pytest.fixture
  def empty_message_parser(self, base_parser):
    return parsers.EmptyMessageParser(base_parser)

  def test_base_parser_parse_returns_none(self, base_parser):
    assert base_parser.parse('') is None

  @pytest.mark.parametrize(
    'element,expected_value',
    [
      (
        NameAttribute('some-name'),
        'some-name',
      ),
      (
        TextAttribute('some-text'),
        'some-text',
      ),
      (
        AssetAttribute('some-asset'),
        'some-asset',
      ),
      (
        ValueAttribute(1),
        1,
      ),
      (
        '',
        None,
      ),
    ],
  )
  def test_attribute_parser_parser_returns_expected_value(
    self, attribute_parser, element, expected_value
  ):
    assert attribute_parser.parse(element) == expected_value

  @pytest.mark.parametrize(
    'element,expected_value',
    [
      (
        FakeMessage(message='test'),
        'Not set',
      ),
      (
        '',
        None,
      ),
    ],
  )
  def test_empty_message_parser_returns_expected_value(
    self, empty_message_parser, element, expected_value
  ):
    assert empty_message_parser.parse(element) == expected_value


class TestGoogleAdsRowParser:
  @pytest.fixture
  def fake_change_event(self):
    return change_event.ChangeEvent.ChangedResource.from_json(
      json.dumps({'campaign': {'target_cpa': {'target_cpa_micros': 1}}})
    )

  @pytest.fixture
  def fake_policy_summary(self):
    return ad_group_ad_asset_view.AdGroupAdAssetPolicySummary.from_json(
      json.dumps(
        {
          'approvalStatus': 'APPROVED',
          'policyTopicEntries': [{'type': 'LIMITED'}],
        }
      )
    )

  @pytest.fixture
  def fake_approval_status(self):
    return ad_group_ad_asset_view.AdGroupAdAssetPolicySummary.from_json(
      json.dumps({'approvalStatus': 'APPROVED'})
    )

  @pytest.fixture
  def fake_ads_row(
    self, fake_change_event, fake_policy_summary, fake_approval_status
  ):
    return FakeAdsRowMultipleElements(
      campaign_type=NameAttribute('SEARCH'),
      clicks=1,
      resource='customers/1/resource/2',
      value=ValueAttribute(1),
      metrics=Metric(clicks=10, impressions=10),
      old_resource=NestedResource('nested_value'),
      new_resource=fake_change_event,
      policy_summary=fake_policy_summary,
    )

  @pytest.fixture
  def fake_expression_virtual_column(self):
    return query_editor.VirtualColumn(
      type='expression',
      value='metrics.clicks / metrics.impressions',
      fields=[
        'metrics.clicks',
        'metrics.impressions',
      ],
      substitute_expression='{metrics_clicks} / {metrics_impressions}',
    )

  def test_google_ads_row_parser_return_last_parser_in_chain(
    self, google_ads_row_parser
  ):
    assert isinstance(
      google_ads_row_parser.parser_chain, parsers.RepeatedParser
    )

  def test_get_attributes_from_row_returns_correct_list(
    self,
    google_ads_row_parser,
    fake_ads_row,
    fake_change_event,
    fake_policy_summary,
  ):
    extracted_rows = google_ads_row_parser._get_attributes_from_row(
      fake_ads_row, google_ads_row_parser.row_getter
    )
    assert extracted_rows == (
      NameAttribute('SEARCH'),
      1,
      'customers/1/resource/2',
      ValueAttribute(1),
      NestedResource('nested_value'),
      fake_change_event,
      fake_policy_summary,
      fake_policy_summary,
    )
    assert google_ads_row_parser.parse_ads_row(fake_ads_row) == [
      'SEARCH',
      1,
      2,
      1,
      'nested_value',
      1,
      ['LIMITED'],
      'APPROVED',
    ]

  def test_parse_ads_row_extracts_correct_resource_indices_from_array(
    self,
    google_ads_row_parser,
    fake_change_event,
    fake_policy_summary,
    fake_approval_status,
  ):
    fake_ads_row_with_array = FakeAdsRowMultipleElements(
      campaign_type=NameAttribute('SEARCH'),
      clicks=1,
      resource=[
        'customers/1/resource/1',
        'customers/1/resource/2',
      ],
      value=ValueAttribute(1),
      metrics=Metric(clicks=10, impressions=10),
      old_resource=NestedResource('nested_value'),
      new_resource=fake_change_event,
      policy_summary=fake_policy_summary,
    )
    assert google_ads_row_parser.parse_ads_row(fake_ads_row_with_array) == [
      'SEARCH',
      1,
      [1, 2],
      1,
      'nested_value',
      1,
      ['LIMITED'],
      'APPROVED',
    ]

  def test_parse_ads_row_extract_correct_resource_indices_from_array_of_attributes(  # pylint: disable=line-too-long
    self,
    google_ads_row_parser,
    fake_change_event,
    fake_policy_summary,
    fake_approval_status,
  ):
    fake_ads_row_with_array = FakeAdsRowMultipleElements(
      campaign_type=NameAttribute('SEARCH'),
      clicks=1,
      resource=[
        AssetAttribute('customers/1/resource/1'),
        AssetAttribute('customers/1/resource/2'),
      ],
      value=ValueAttribute(1),
      metrics=Metric(clicks=10, impressions=10),
      old_resource=NestedResource('nested_value'),
      new_resource=fake_change_event,
      policy_summary=fake_policy_summary,
    )
    assert google_ads_row_parser.parse_ads_row(fake_ads_row_with_array) == [
      'SEARCH',
      1,
      [1, 2],
      1,
      'nested_value',
      1,
      ['LIMITED'],
      'APPROVED',
    ]

  def test_convert_virtual_column_returns_correct_value_for_builtin_type(
    self, google_ads_row_parser, fake_ads_row
  ):
    fake_builtin_virtual_column = query_editor.VirtualColumn(
      type='built-in', value='fake_value'
    )
    result = google_ads_row_parser._convert_virtual_column(
      fake_ads_row, fake_builtin_virtual_column
    )
    assert result == 'fake_value'

  def test_convert_virtual_column_returns_correct_value_for_expression(
    self, google_ads_row_parser, fake_ads_row, fake_expression_virtual_column
  ):
    result = google_ads_row_parser._convert_virtual_column(
      fake_ads_row, fake_expression_virtual_column
    )
    assert result == 1.0

  def test_convert_virtual_column_returns_zero_for_expression_with_zero_in_denominator(  # pylint: disable=line-too-long
    self, google_ads_row_parser, fake_ads_row, fake_expression_virtual_column
  ):
    fake_ads_row.metrics.impressions = 0
    result = google_ads_row_parser._convert_virtual_column(
      fake_ads_row, fake_expression_virtual_column
    )
    assert result == 0

  def test_convert_virtual_column_raises_virtual_column_exception_on_incorrect_type(  # pylint: disable=line-too-long
    self, google_ads_row_parser, fake_ads_row, fake_expression_virtual_column
  ):
    fake_ads_row.metrics.impressions = 'str'
    with pytest.raises(exceptions.GaarfVirtualColumnException):
      google_ads_row_parser._convert_virtual_column(
        fake_ads_row, fake_expression_virtual_column
      )

  def test_convert_virtual_column_returns_virtual_column_value_on_incorrect_expression(  # pylint: disable=line-too-long
    self, google_ads_row_parser, fake_ads_row, fake_expression_virtual_column
  ):
    fake_ads_row.metrics.impressions = '0 +'
    result = google_ads_row_parser._convert_virtual_column(
      fake_ads_row, fake_expression_virtual_column
    )
    assert result == fake_expression_virtual_column.value

  def test_convert_virtual_column_raises_virtual_column_exception_on_incorrect_type(  # pylint: disable=line-too-long
    self, google_ads_row_parser, fake_ads_row
  ):
    virtual_column = query_editor.VirtualColumn(
      type='non-existing-type',
      value='metrics.clicks / metrics.impressions',
      fields=[
        'metrics.clicks',
        'metrics.impressions',
      ],
      substitute_expression='{metrics_clicks} / {metrics_impressions}',
    )
    with pytest.raises(exceptions.GaarfVirtualColumnException):
      google_ads_row_parser._convert_virtual_column(
        fake_ads_row, virtual_column
      )

  def test_parse_ads_row_with_extract_protobufs_returns_correct_results(
    self, google_ads_row_parser, fake_ads_row
  ):
    google_ads_row_parser.extract_protobufs = True
    assert google_ads_row_parser.parse_ads_row(fake_ads_row) == [
      'SEARCH',
      1,
      2,
      1,
      'nested_value',
      1,
      ['LIMITED'],
      'APPROVED',
    ]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for executing Gaarf queries and writing them to local/remote.

AdsQueryExecutor performs fetching data from Ads API in a form of
GaarfReport and saving it to local/remote storage.
"""

from __future__ import annotations

import logging
from collections.abc import MutableSequence

from google.ads.googleads import client as googleads_client

from gaarf import api_clients, query_editor, report_fetcher
from gaarf.io.writers import abs_writer, console_writer

logger = logging.getLogger(__name__)


class AdsQueryExecutor:
  """Gets data from Ads API and writes them to local/remote storage.

  Attributes:
      api_client: a client used for connecting to Ads API.
  """

  def __init__(
    self,
    api_client: api_clients.GoogleAdsApiClient
    | googleads_client.GoogleAdsClient,
  ) -> None:
    """Initializes QueryExecutor.

    Args:
        api_client: a client used for connecting to Ads API.
    """
    self.api_client = api_client

  @property
  def report_fetcher(self) -> report_fetcher.AdsReportFetcher:
    """Initializes AdsReportFetcher to get data from Ads API."""
    return report_fetcher.AdsReportFetcher(self.api_client)

  def execute(
    self,
    query_text: str,
    query_name: str,
    customer_ids: list[str] | str,
    writer_client: abs_writer.AbsWriter = console_writer.ConsoleWriter(),
    args: dict[str, str] | None = None,
    optimize_performance: str = 'NONE',
  ) -> None:
    """Reads query, extract results and stores them in a specified location.

    Args:
        query_text: Text for the query.
        query_name: Identifier of a query..
        customer_ids: All accounts for which query will be executed.
        writer_client: Client responsible for writing data to local/remote
            location.
        args: Arguments that need to be passed to the query
        optimize_performance: strategy for speeding up query execution
            ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF")
    """
    query_specification = query_editor.QuerySpecification(
      query_text, query_name, args, self.report_fetcher.api_client.api_version
    ).generate()
    results = self.report_fetcher.fetch(
      query_specification=query_specification,
      customer_ids=customer_ids,
      optimize_strategy=optimize_performance,
    )
    logger.debug(
      'Start writing data for query %s via %s writer',
      query_specification.query_title,
      type(writer_client),
    )
    writer_client.write(results, query_specification.query_title)
    logger.debug(
      'Finish writing data for query %s via %s writer',
      query_specification.query_title,
      type(writer_client),
    )

  def expand_mcc(
    self,
    customer_ids: str | MutableSequence,
    customer_ids_query: str | None = None,
  ) -> list[str]:
    """Performs Manager account(s) expansion to child accounts.

    Args:
        customer_ids:
            Manager account(s) to be expanded.
        customer_ids_query:
            Gaarf query to limit the expansion only to accounts
            satisfying the condition.

    Returns:
        All child accounts under provided customer_ids.
    """
    return self.report_fetcher.expand_mcc(customer_ids, customer_ids_query)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Dict, Optional

from jinja2 import Environment, FileSystemLoader, Template

logger = logging.getLogger(__name__)


class PostProcessorMixin:
  def replace_params_template(
    self, query_text: str, params: Optional[Dict[str, Any]] = None
  ) -> str:
    logger.debug('Original query text:\n%s', query_text)
    if params:
      if templates := params.get('template'):
        query_templates = {
          name: value for name, value in templates.items() if name in query_text
        }
        if query_templates:
          query_text = self.expand_jinja(query_text, query_templates)
          logger.debug('Query text after jinja expansion:\n%s', query_text)
        else:
          query_text = self.expand_jinja(query_text, {})
      else:
        query_text = self.expand_jinja(query_text, {})
      if macros := params.get('macro'):
        query_text = query_text.format(**macros)
        logger.debug('Query text after macro substitution:\n%s', query_text)
    else:
      query_text = self.expand_jinja(query_text, {})
    return query_text

  def expand_jinja(
    self, query_text: str, template_params: Optional[Dict[str, Any]] = None
  ) -> str:
    file_inclusions = ('% include', '% import', '% extend')
    if any(file_inclusion in query_text for file_inclusion in file_inclusions):
      template = Environment(loader=FileSystemLoader('.'))
      query = template.from_string(query_text)
    else:
      query = Template(query_text)
    if not template_params:
      return query.render()
    for key, value in template_params.items():
      if value:
        if isinstance(value, list):
          template_params[key] = value
        elif len(splitted_param := value.split(',')) > 1:
          template_params[key] = splitted_param
        else:
          template_params[key] = value
      else:
        template_params = ''
    return query.render(template_params)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defing `gaarf` CLI utility.

`gaarf` allows to execute GAQL queries and store results in local/remote
storage.
"""

from __future__ import annotations

import argparse
import functools
from collections.abc import MutableSequence
from concurrent import futures
from pathlib import Path

import smart_open
import yaml

from gaarf import api_clients, exceptions, query_executor
from gaarf.cli import utils
from gaarf.io import reader, writer


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('query', nargs='*')
  parser.add_argument('-c', '--config', dest='gaarf_config', default=None)
  parser.add_argument('--account', dest='account', default=None)
  parser.add_argument('--output', dest='output', default=None)
  parser.add_argument('--input', dest='input', default='file')
  parser.add_argument(
    '--ads-config', dest='config', default=str(Path.home() / 'google-ads.yaml')
  )
  parser.add_argument('--api-version', dest='api_version', default=None)
  parser.add_argument('--log', '--loglevel', dest='loglevel', default='info')
  parser.add_argument('--logger', dest='logger', default='local')
  parser.add_argument(
    '--customer-ids-query', dest='customer_ids_query', default=None
  )
  parser.add_argument(
    '--customer-ids-query-file', dest='customer_ids_query_file', default=None
  )
  parser.add_argument('--save-config', dest='save_config', action='store_true')
  parser.add_argument(
    '--no-save-config', dest='save_config', action='store_false'
  )
  parser.add_argument(
    '--config-destination', dest='save_config_dest', default='config.yaml'
  )
  parser.add_argument(
    '--parallel-queries', dest='parallel_queries', action='store_true'
  )
  parser.add_argument(
    '--no-parallel-queries', dest='parallel_queries', action='store_false'
  )
  parser.add_argument(
    '--optimize-performance', dest='optimize_performance', default='NONE'
  )
  parser.add_argument('--dry-run', dest='dry_run', action='store_true')
  parser.add_argument(
    '--disable-account-expansion',
    dest='disable_account_expansion',
    action='store_true',
  )
  parser.add_argument('-v', '--version', dest='version', action='store_true')
  parser.add_argument(
    '--parallel-threshold', dest='parallel_threshold', default=None
  )
  parser.set_defaults(save_config=False)
  parser.set_defaults(parallel_queries=True)
  parser.set_defaults(dry_run=False)
  parser.set_defaults(disable_account_expansion=False)
  args = parser.parse_known_args()
  main_args = args[0]

  if main_args.version:
    import pkg_resources

    version = pkg_resources.require('google-ads-api-report-fetcher')[0].version
    print(f'gaarf version {version}')
    exit()

  logger = utils.init_logging(
    loglevel=main_args.loglevel.upper(), logger_type=main_args.logger
  )
  if not main_args.query:
    logger.error('Please provide one or more queries to run')
    raise exceptions.GaarfMissingQueryException(
      'Please provide one or more queries to run'
    )

  with smart_open.open(main_args.config, 'r', encoding='utf-8') as f:
    google_ads_config_dict = yaml.safe_load(f)

  config = utils.ConfigBuilder('gaarf').build(vars(main_args), args[1])
  if not config.account:
    if mcc := google_ads_config_dict.get('login_customer_id'):
      config.account = str(mcc)
    else:
      raise exceptions.GaarfMissingAccountException(
        'No account found, please specify via --account CLI flag'
        'or add as login_customer_id in google-ads.yaml'
      )
  logger.debug('config: %s', config)

  if main_args.save_config and not main_args.gaarf_config:
    utils.ConfigSaver(main_args.save_config_dest).save(config)
  if main_args.dry_run:
    exit()

  if config.params:
    config = utils.initialize_runtime_parameters(config)
  logger.debug('initialized config: %s', config)

  ads_client = api_clients.GoogleAdsApiClient(
    config_dict=google_ads_config_dict,
    version=config.api_version,
    use_proto_plus=main_args.optimize_performance
    not in ('PROTOBUF', 'BATCH_PROTOBUF'),
  )
  ads_query_executor = query_executor.AdsQueryExecutor(ads_client)
  reader_factory = reader.ReaderFactory()
  reader_client = reader_factory.create_reader(main_args.input)

  if config.customer_ids_query:
    customer_ids_query = config.customer_ids_query
  elif config.customer_ids_query_file:
    file_reader = reader_factory.create_reader('file')
    customer_ids_query = file_reader.read(config.customer_ids_query_file)
  else:
    customer_ids_query = None

  if main_args.disable_account_expansion:
    logger.info(
      'Skipping account expansion because of ' 'disable_account_expansion flag'
    )
    customer_ids = (
      config.account
      if isinstance(config.account, MutableSequence)
      else [config.account]
    )
  else:
    customer_ids = ads_query_executor.expand_mcc(
      config.account, customer_ids_query
    )
  if not customer_ids:
    logger.warning(
      'Not a single under MCC %s is found that satisfies '
      'the following customer_id query: "%s"',
      config.account,
      customer_ids_query,
    )
    exit()
  writer_client = writer.WriterFactory().create_writer(
    config.output, **config.writer_params
  )
  if config.output == 'bq':
    _ = writer_client.create_or_get_dataset()
  if config.output == 'sheet':
    writer_client.init_client()

  logger.info(
    'Total number of customer_ids is %d, accounts=[%s]',
    len(customer_ids),
    ','.join(map(str, customer_ids)),
  )

  if main_args.parallel_queries:
    logger.info('Running queries in parallel')
    with futures.ThreadPoolExecutor(main_args.parallel_threshold) as executor:
      future_to_query = {
        executor.submit(
          ads_query_executor.execute,
          reader_client.read(query),
          query,
          customer_ids,
          writer_client,
          config.params,
          main_args.optimize_performance,
        ): query
        for query in main_args.query
      }
      for future in futures.as_completed(future_to_query):
        query = future_to_query[future]
        utils.gaarf_runner(query, future.result, logger)
  else:
    logger.info('Running queries sequentially')
    for query in main_args.query:
      callback = functools.partial(
        ads_query_executor.execute,
        reader_client.read(query),
        query,
        customer_ids,
        writer_client,
        config.params,
        main_args.optimize_performance,
      )
      utils.gaarf_runner(query, callback, logger)


if __name__ == '__main__':
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains helpers classes to simulate Google Ads API response."""

from __future__ import annotations

import dataclasses
from typing import Generator

from gaarf import parsers


@dataclasses.dataclass
class FakeResponse:
  """Simulates Google Ads API response iterator."""

  data: list[list[parsers.GoogleAdsRowElement]]

  def __iter__(self) -> Generator[FakeBatch, None, None]:
    """Yields batch from Google Ads API response."""
    for result in self.data:
      yield FakeBatch(result)


@dataclasses.dataclass
class FakeBatch:
  """Simulates a single batch of data from Google Ads API response.

  Attributes:
    results: Data from API response.
    query_resource_consumption: Resources consumed to serve the query.
  """

  results: list[parsers.GoogleAdsRowElement]
  query_resource_consumption: int = 0


@dataclasses.dataclass
class Customer:
  """Helper to represent customer resource."""

  id: int


@dataclasses.dataclass
class FakeGoogleAdsRowElement:
  """Helper to represent a single row in the batch."""

  customer: Customer



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines mechanism for executing queries via SqlAlchemy."""

from __future__ import annotations

import warnings

import sqlalchemy

from gaarf.executors import sql_executor


class SqlAlchemyQueryExecutor:
  """Deprecated class for creating SqlAlchemyQueryExecutor."""

  def __new__(
    cls, engine: sqlalchemy.engine.base.Engine
  ) -> sql_executor.SqlAlchemyQueryExecutor:
    warnings.warn(
      'Loading SqlAlchemyQueryExecutor from `gaarf.sql_executor` is '
      'deprecated; Import SqlAlchemyQueryExecutor from '
      '`gaarf.executors.sql_executor` instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return sql_executor.SqlAlchemyQueryExecutor(engine)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data to BigQuery."""

from __future__ import annotations

try:
  from google.cloud import bigquery
except ImportError as e:
  raise ImportError(
    'Please install google-ads-api-report-fetcher with BigQuery support '
    '- `pip install google-ads-api-report-fetcher[bq]`'
  ) from e

import datetime
import logging
from collections.abc import Sequence

import numpy as np
import pandas as pd
import proto  # type: ignore
from google.cloud import exceptions as google_cloud_exceptions

import gaarf
from gaarf import parsers
from gaarf.io import formatter
from gaarf.io.writers import abs_writer


class BigQueryWriter(abs_writer.AbsWriter):
  """Writes Gaarf Report to BigQuery.

  Attributes:
    project: Id of Google Cloud Project.
    dataset: BigQuery dataset to write data to.
    location: Location of a newly created dataset.
    write_disposition: Option for overwriting data.
  """

  def __init__(
    self,
    project: str,
    dataset: str,
    location: str = 'US',
    write_disposition: bigquery.WriteDisposition = bigquery.WriteDisposition.WRITE_TRUNCATE,
    **kwargs,
  ):
    """Initializes BigQueryWriter.

    Args:
      project: Id of Google Cloud Project.
      dataset: BigQuery dataset to write data to.
      location: Location of a newly created dataset.
      write_disposition: Option for overwriting data.
      kwargs: Optional keywords arguments.
    """
    super().__init__(**kwargs)
    self.project = project
    self.dataset_id = f'{project}.{dataset}'
    self.location = location
    self.write_disposition = write_disposition

  def __str__(self) -> str:
    return f'[BigQuery] - {self.dataset_id} at {self.location} location.'

  @property
  def client(self) -> bigquery.Client:
    """Instantiated BigQuery client."""
    return bigquery.Client(self.project)

  def create_or_get_dataset(self) -> bigquery.Dataset:
    """Gets existing dataset or create a new one."""
    try:
      bq_dataset = self.client.get_dataset(self.dataset_id)
    except google_cloud_exceptions.NotFound:
      bq_dataset = bigquery.Dataset(self.dataset_id)
      bq_dataset.location = self.location
      bq_dataset = self.client.create_dataset(bq_dataset, timeout=30)
    return bq_dataset

  def write(self, report: gaarf.report.GaarfReport, destination: str) -> str:
    """Writes Gaarf report to a BigQuery table.

    Args:
      report: Gaarf report.
      destination: Name of the table report should be written to.

    Returns:
      Name of the table in `dataset.table` format.
    """
    report = self.format_for_write(report)
    schema = _define_schema(report)
    destination = formatter.format_extension(destination)
    table = self._create_or_get_table(
      f'{self.dataset_id}.{destination}', schema
    )
    job_config = bigquery.LoadJobConfig(
      write_disposition=self.write_disposition,
      schema=schema,
      source_format='CSV',
      max_bad_records=len(report),
    )

    if not report:
      df = pd.DataFrame(
        data=report.results_placeholder, columns=report.column_names
      ).head(0)
    else:
      df = report.to_pandas()
    df = df.replace({np.nan: None})
    logging.debug('Writing %d rows of data to %s', len(df), destination)
    job = self.client.load_table_from_dataframe(
      dataframe=df, destination=table, job_config=job_config
    )
    try:
      job.result()
      logging.debug('Writing to %s is completed', destination)
    except google_cloud_exceptions.BadRequest as e:
      raise ValueError(f'Unable to save data to BigQuery! {str(e)}') from e
    return f'[BigQuery] - at {self.dataset_id}.{destination}'

  def _create_or_get_table(
    self, table_name: str, schema: Sequence[bigquery.SchemaField]
  ) -> bigquery.Table:
    """Gets existing table or create a new one.

    Args:
      table_name: Name of the table in BigQuery.
      schema: Schema of the table if one should be created.

    Returns:
      BigQuery table object.
    """
    try:
      table = self.client.get_table(table_name)
    except google_cloud_exceptions.NotFound:
      table_ref = bigquery.Table(table_name, schema=schema)
      table = self.client.create_table(table_ref)
      table = self.client.get_table(table_name)
    return table


def _define_schema(
  report: gaarf.report.GaarfReport,
) -> list[bigquery.SchemaField]:
  """Infers schema from GaarfReport.

  Args:
    report: GaarfReport to infer schema from.

  Returns:
    Schema fields for a given report.

  """
  result_types = _get_result_types(report)
  return _get_bq_schema(result_types)


def _get_result_types(
  report: gaarf.report.GaarfReport,
) -> dict[str, dict[str, parsers.GoogleAdsRowElement]]:
  """Maps each column of report to BigQuery field type and repeated status.

  Fields types are inferred based on report resuls or results placeholder.

  Args:
    report: GaarfReport to infer field types from.

  Returns:
    Mapping between each column of report and its field type.
  """
  result_types: dict[str, dict[str, parsers.GoogleAdsRowElement]] = {}
  column_names = report.column_names
  for row in report.results or report.results_placeholder:
    if set(column_names) == set(result_types.keys()):
      break
    for i, field in enumerate(row):
      if field is None or column_names[i] in result_types:
        continue
      field_type = type(field)
      if field_type in [
        list,
        proto.marshal.collections.repeated.RepeatedComposite,
        proto.marshal.collections.repeated.Repeated,
      ]:
        repeated = True
        if len(field) == 0:
          field_type = str
        else:
          field_type = type(field[0])
      else:
        field_type = type(field)
        repeated = False
      result_types[column_names[i]] = {
        'field_type': field_type,
        'repeated': repeated,
      }
  return result_types


def _get_bq_schema(
  types: dict[str, dict[str, parsers.GoogleAdsRowElement]],
) -> list[bigquery.SchemaField]:
  """Converts GoogleAds fields types to BigQuery schema fields.

  Args:
    types: Mapping between column names and its field type.

  Returns:
     BigQuery schema fields corresponding to GaarfReport.
  """
  type_mapping = {
    list: 'REPEATED',
    str: 'STRING',
    datetime.datetime: 'DATETIME',
    datetime.date: 'DATE',
    int: 'INT64',
    float: 'FLOAT64',
    bool: 'BOOL',
    proto.marshal.collections.repeated.RepeatedComposite: 'REPEATED',
    proto.marshal.collections.repeated.Repeated: 'REPEATED',
  }

  schema: list[bigquery.SchemaField] = []
  for key, value in types.items():
    field_type = type_mapping.get(value.get('field_type'))
    schema.append(
      bigquery.SchemaField(
        name=key,
        field_type=field_type if field_type else 'STRING',
        mode='REPEATED' if value.get('repeated') else 'NULLABLE',
      )
    )
  return schema



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
from __future__ import annotations

import dataclasses

import pytest

from gaarf import api_clients, report_fetcher, utils
from tests.unit import helpers


@dataclasses.dataclass
class CustomerClient:
  id: int


@dataclasses.dataclass
class FakeGoogleAdsRowElement:
  customer_client: CustomerClient


@pytest.fixture
def test_client(mocker, config_path):
  fake_results = [
    [
      FakeGoogleAdsRowElement(CustomerClient(1)),
    ],
  ]
  fake_response = helpers.FakeResponse(data=fake_results)
  mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
  mocker.patch(
    'gaarf.api_clients.GoogleAdsApiClient.get_response',
    return_value=fake_response,
  )
  return api_clients.GoogleAdsApiClient(path_to_config=config_path)


@pytest.fixture
def fake_report_fetcher(mocker, test_client):
  data = ['1']
  mocker.patch(
    'gaarf.report_fetcher.AdsReportFetcher.expand_mcc', return_value=data
  )
  return report_fetcher.AdsReportFetcher(test_client)


def test_calling_get_customer_ids_is_deprecated(test_client):
  with pytest.warns(DeprecationWarning) as w:
    utils.get_customer_ids(test_client, '1')
    assert len(w) == 1
    assert str(w[0].message) == (
      '`get_customer_ids` will be deprecated, '
      'use `AdsReportFetcher.expand_mcc` or '
      '`AdsQueryExecutor.expand_mcc` '
      'methods instead'
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for various utility functions."""

from __future__ import annotations

import warnings
from collections.abc import MutableSequence

from gaarf import api_clients, report_fetcher


def get_customer_ids(
  ads_client: api_clients.GoogleAdsApiClient,
  customer_id: str | MutableSequence,
  customer_ids_query: str | None = None,
) -> list[str]:
  """Gets list of customer_ids from an MCC account.

  Args:
      ads_client: GoogleAdsApiClient used for connection.
      customer_id: MCC account_id(s).
      customer_ids_query: GAQL query used to reduce the number of accounts.

  Returns:
      All customer_ids from MCC satisfying the condition.
  """
  warnings.warn(
    '`get_customer_ids` will be deprecated, '
    'use `AdsReportFetcher.expand_mcc` or `AdsQueryExecutor.expand_mcc` '
    'methods instead',
    category=DeprecationWarning,
    stacklevel=3,
  )
  return report_fetcher.AdsReportFetcher(ads_client).expand_mcc(
    customer_id, customer_ids_query
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
from __future__ import annotations

import json
import os

import pytest

from gaarf import api_clients
from gaarf.executors import ads_executor
from gaarf.io.writers import json_writer
from tests.unit import helpers


class TestAdsQueryExecutor:
  @pytest.fixture
  def test_client(self, mocker, config_path):
    mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
    return api_clients.GoogleAdsApiClient(path_to_config=config_path)

  @pytest.fixture
  def fake_response(self):
    fake_results = [
      [
        helpers.FakeGoogleAdsRowElement(helpers.Customer(1)),
      ],
      [
        helpers.FakeGoogleAdsRowElement(helpers.Customer(2)),
      ],
      [
        helpers.FakeGoogleAdsRowElement(helpers.Customer(3)),
      ],
    ]
    return helpers.FakeResponse(data=fake_results)

  @pytest.fixture
  def executor(self, mocker, test_client, fake_response):
    mocker.patch(
      'gaarf.api_clients.GoogleAdsApiClient.get_response',
      return_value=fake_response,
    )
    return ads_executor.AdsQueryExecutor(test_client)

  @pytest.fixture
  def test_json_writer(self, tmp_path):
    return json_writer.JsonWriter(destination_folder=tmp_path)

  def test_execute_returns_success(self, executor, test_json_writer):
    query_text = 'SELECT customer.id FROM customer'
    expected_result = [
      {'customer_id': 1},
      {'customer_id': 2},
      {'customer_id': 3},
    ]

    executor.execute(
      query_text=query_text,
      query_name='test',
      customer_ids='1234567890',
      writer_client=test_json_writer,
    )
    with open(
      os.path.join(test_json_writer.destination_folder, 'test.json'),
      'r',
      encoding='utf-8',
    ) as f:
      result = json.load(f)

    assert result == expected_result



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining various parsing strategy for GoogleAdsRow elements.

GoogleAdsRowParser parses a single GoogleAdsRow and applies different parsing
strategies to each element of the row.
"""

from __future__ import annotations

import ast
import contextlib
import importlib
import operator
import re
from collections import abc
from typing import Union, get_args

import proto  # type: ignore
from google import protobuf
from google.ads.googleads import util as googleads_utils
from proto.marshal.collections import repeated
from typing_extensions import Self, TypeAlias

from gaarf import api_clients, exceptions, query_editor

google_ads_service = importlib.import_module(
  f'google.ads.googleads.{api_clients.GOOGLE_ADS_API_VERSION}.'
  'services.types.google_ads_service'
)

GoogleAdsRowElement: TypeAlias = Union[int, float, str, bool, list, None]

_REPEATED: TypeAlias = Union[
  repeated.Repeated,
  protobuf.internal.containers.RepeatedScalarFieldContainer,
]
_REPEATED_COMPOSITE: TypeAlias = Union[
  repeated.RepeatedComposite,
  protobuf.internal.containers.RepeatedCompositeFieldContainer,
]

_NESTED_FIELD: TypeAlias = Union[
  _REPEATED,
  _REPEATED_COMPOSITE,
]


class BaseParser:
  """Base class for defining parsers.

  Attributes:
      _successor: Indicates the previous parser in the chain.
  """

  def __init__(self, successor: type[Self]) -> None:
    self._successor = successor

  def parse(self, element: GoogleAdsRowElement) -> GoogleAdsRowElement:
    """Parses GoogleAdsRow by using a successor parser.

    Args:
        element: An element of a GoogleAdsRow.

    Returns:
        Parsed GoogleAdsRow element.
    """
    if self._successor:
      return self._successor.parse(element)
    return None


class RepeatedParser(BaseParser):
  """Parses repeated.Repeated resources."""

  def parse(self, element: GoogleAdsRowElement) -> GoogleAdsRowElement:
    """Parses only repeated elements from GoogleAdsRow.

    If there a repeated resource, applies transformations to each element;
    otherwise delegates parsing of the element to the next parser
    in the chain.

    Args:
        element: An element of a GoogleAdsRow.

    Returns:
        Parsed GoogleAdsRow element.
    """
    if isinstance(element, get_args(_REPEATED)) and 'customer' in str(element):
      items: list[GoogleAdsRowElement] = []
      for item in element:
        items.append(
          ResourceFormatter(item).get_resource_id().clean_resource_id().format()
        )
      return items
    return super().parse(element)


class RepeatedCompositeParser(BaseParser):
  """Parses repeated.RepeatedComposite elements."""

  def parse(self, element):
    """Parses only repeated composite resources from GoogleAdsRow.

    If there a repeated composited resource, applies transformations
    to each element; otherwise delegates parsing of the element
    to the next parser in the chain.

    Args:
        element: An element of a GoogleAdsRow.

    Returns:
        Parsed GoogleAdsRow element.
    """
    if isinstance(element, get_args(_REPEATED_COMPOSITE)):
      items = []
      for item in element:
        items.append(
          ResourceFormatter(item)
          .get_nested_resource()
          .get_resource_id()
          .clean_resource_id()
          .format()
        )
      return items
    return super().parse(element)


class AttributeParser(BaseParser):
  """Parses elements that have attributes."""

  def parse(self, element: GoogleAdsRowElement) -> GoogleAdsRowElement:
    """Parses only elements that have attributes.

    If there a repeated composited resource, applies transformations
    to each element; otherwise delegates parsing of the element
    to the next parser in the chain.

    Args:
        element: An element of a GoogleAdsRow.
    Returns:
        Parsed GoogleAdsRow element.
    """
    if hasattr(element, 'name'):
      return element.name
    if hasattr(element, 'text'):
      return element.text
    if hasattr(element, 'asset'):
      return element.asset
    if hasattr(element, 'value'):
      return element.value
    return super().parse(element)


class EmptyMessageParser(BaseParser):
  """Generates placeholder for empty Message objects."""

  def parse(self, element: GoogleAdsRowElement) -> GoogleAdsRowElement:
    """Checks if an element is an empty proto.Message.

    If an element is empty message, returns 'Not set' placeholder;
    otherwise delegates parsing of the element to the next parser
    in the chain.

    Args:
        element: An element of a GoogleAdsRow.
    Returns:
        Parsed GoogleAdsRow element.
    """
    if issubclass(type(element), proto.Message):
      return 'Not set'
    return super().parse(element)


class GoogleAdsRowParser:
  """Performs parsing of a single GoogleAdsRow.

  Attributes:
      fields: Expected fields in GoogleAdsRow.
      customizers: Customizing behaviour performed on a field.
      virtual_columns: Elements that are not directly present in GoogleAdsRow.
      parser: Chain of parsers to execute on a single GoogleAdsRow.
      row_getter: Helper to easily extract fields from GogleAdsRow.
      respect_nulls: Whether or not convert nulls to zeros.
  """

  def __init__(self, query_specification: query_editor.QueryElements) -> None:
    """Initializes GoogleAdsRowParser.

    Args:
        query_specification: All elements forming gaarf query.
    """
    self.fields = query_specification.fields
    self.customizers = query_specification.customizers
    self.virtual_columns = query_specification.virtual_columns
    self.column_names = query_specification.column_names
    self.parser_chain = self._init_parsers_chain()
    self.row_getter = operator.attrgetter(*query_specification.fields)
    # Some segments are automatically converted to 0 when not present
    # For this case we specify attribute `respect_null` which converts
    # such attributes to None rather than 0
    self.respect_nulls = (
      'segments.sk_ad_network_conversion_value' in self.fields
    )

  def _init_parsers_chain(self):
    """Initializes chain of parsers."""
    parser_chain = BaseParser(None)
    for parser in [
      EmptyMessageParser,
      AttributeParser,
      RepeatedCompositeParser,
      RepeatedParser,
    ]:
      new_parser = parser(parser_chain)
      parser_chain = new_parser
    return parser_chain

  def parse_ads_row(
    self, row: google_ads_service.GoogleAdsRow
  ) -> list[GoogleAdsRowElement]:
    """Parses GoogleAdsRow by applying various transformations.

    Args:
        row: A single GoogleAdsRow.

    Returns:
        List of parsed elements.
    """
    parsed_row_elements: list[GoogleAdsRowElement] = []
    extracted_attributes = self._get_attributes_from_row(row, self.row_getter)
    index = 0
    for column in self.column_names:
      if column in self.virtual_columns.keys():
        parsed_element = self._convert_virtual_column(
          row, self.virtual_columns[column]
        )
      else:
        parsed_element = self._parse_row_element(
          extracted_attributes[index], column
        )
        index += 1
      parsed_row_elements.append(parsed_element)
    return parsed_row_elements

  def _parse_row_element(
    self, extracted_attribute: GoogleAdsRowElement, column: str
  ) -> GoogleAdsRowElement:
    """Parses a single element from row.

    Args:
        extracted_attribute: A single element from GoogleAdsRow.
        column: Corresponding name of the element.

    Returns:
        Parsed element.
    """
    if self.customizers:
      extracted_attribute = self._extract_attributes_with_customizer(
        extracted_attribute, column
      )
    if isinstance(extracted_attribute, abc.MutableSequence):
      parsed_element = [
        self.parser_chain.parse(element) or element
        for element in extracted_attribute
      ]
    else:
      parsed_element = (
        self.parser_chain.parse(extracted_attribute) or extracted_attribute
      )

    return parsed_element

  def _extract_attributes_with_customizer(
    self, extracted_attribute: GoogleAdsRowElement, column: str
  ) -> GoogleAdsRowElement:
    """Extracts additional info from row element based on customizers.

    Some GoogleAdsRow objects can be complex and customizers help extract
    specific values from them by using special syntax.

    Args:
        extracted_attribute: A single element from GoogleAdsRow.
        column: Corresponding name of the element.

    Returns:
        Extracted row attribute.
    """
    if caller := self.customizers.get(column):
      if caller.get('type') == 'nested_field':
        extracted_attribute = self._extract_nested_customizer(
          extracted_attribute, caller
        )
      elif caller.get('type') == 'resource_index':
        extracted_attribute = self._get_resource_index(
          extracted_attribute, caller
        )
    return extracted_attribute

  def _extract_nested_customizer(
    self, extracted_attribute, caller: dict[str, str]
  ) -> GoogleAdsRowElement:
    """Extracts additional info from nested resource.

    Some GoogleAdsRow objects are nested and has attributes that can be
    further accessed using special customizer syntax.

    Args:
        extracted_attribute: A single element from GoogleAdsRow.
        caller: Mapping between type of customizer type and its value.

    Returns:
        Extracted row attribute.

    Raises:
        GaarfCustomizerException: When customizer incorrectly specified.
    """
    values_ = caller.get('value').split('.')
    extracted_attribute_ = (
      getattr(extracted_attribute, values_[0])
      if hasattr(extracted_attribute, values_[0])
      else extracted_attribute
    )
    try:
      if isinstance(extracted_attribute, get_args(_NESTED_FIELD)) or isinstance(
        extracted_attribute_, get_args(_NESTED_FIELD)
      ):
        if isinstance(
          extracted_attribute_, (repeated.Repeated, repeated.RepeatedComposite)
        ):
          extracted_attribute = extracted_attribute
        if len(values_) > 1:
          value = values_[1]
        else:
          value = caller.get('value')
        extracted_attribute = list(
          {
            operator.attrgetter(value)(element)
            for element in extracted_attribute_
          }
        )
      else:
        extracted_attribute = operator.attrgetter(caller.get('value'))(
          extracted_attribute
        )
      return extracted_attribute
    except AttributeError as e:
      raise exceptions.GaarfCustomizerException(
        f'customizer "{caller}" is incorrect,\n' f'details: "{e}"'
      )

  def _get_resource_index(
    self, extracted_attribute: GoogleAdsRowElement, caller: dict[str, str]
  ) -> int | str:
    """Extracts additional info from resource_name.

    Some GoogleAdsRow objects resource_names
    (i.e. customers/1/conversionActions/2~3); with resource_index we can
    access only the last element of this expression ('3').
    Whenever possible result is converted to integer.

    Args:
        extracted_attribute: A single element from GoogleAdsRow.
        caller: Mapping between type of customizer type and its value.

    Returns:
        Extracted row attribute.
    """
    if isinstance(extracted_attribute, abc.MutableSequence):
      parsed_element = [
        self.parser_chain.parse(element) or element
        for element in extracted_attribute
      ]
      return [
        self._get_resource_index(attribute, caller)
        for attribute in parsed_element
      ]
    extracted_attribute = re.split('~', extracted_attribute)[
      caller.get('value')
    ]
    result = re.split('/', extracted_attribute)[-1]
    try:
      return int(result)
    except ValueError:
      return result

  def _get_attributes_from_row(
    self, row: google_ads_service.GoogleAdsRow, getter: operator.attrgetter
  ) -> tuple[GoogleAdsRowElement, ...]:
    """Extracts attributes from GoogleAdsRow based on attribute getter.

    Attribute getter contains nested attribute access operations (i.e.
    campaign -> id) which can be easily executed on nested GoogleAdsRow
    object.

    Args:
        row: A single GoogleAdsRow.
        getter: Initialized attribute getter.

    Returns:
        All parsed elements from a single GoogleAdsRow..
    """
    attributes = getter(row)
    if self.respect_nulls:
      row = googleads_utils.convert_proto_plus_to_protobuf(row)
      if row.segments.HasField('sk_ad_network_conversion_value'):
        # Convert to list to perform modification
        attributes = list(attributes)
        # Replace 0 attributes in the row with None
        attributes[
          self.fields.index('segments.sk_ad_network_conversion_value')
        ] = None
        # Convert back to tuple
        attributes = tuple(attributes)
    else:
      attributes = getter(row)
    return attributes if isinstance(attributes, tuple) else (attributes,)

  def _convert_virtual_column(
    self,
    row: google_ads_service.GoogleAdsRow,
    virtual_column: query_editor.VirtualColumn,
  ) -> GoogleAdsRowElement:
    """Convert virtual column definition to a single element.

    Args:
        row: A single GoogleAdsRow.
        virtual_column: Virtual column definition.

    Returns:
        Parsed element.
    """
    if virtual_column.type not in ('built-in', 'expression'):
      raise exceptions.GaarfVirtualColumnException(
        f'Unsupported virtual column type: {virtual_column.type}'
      )
    if virtual_column.type == 'built-in':
      return virtual_column.value
    virtual_column_getter = operator.attrgetter(*virtual_column.fields)
    virtual_column_values = virtual_column_getter(row)
    try:
      iter(virtual_column_values)
    except TypeError:
      virtual_column_values = (virtual_column_values,)
    virtual_column_replacements = {
      field.replace('.', '_'): value
      for field, value in zip(virtual_column.fields, virtual_column_values)
    }
    try:
      virtual_column_expression = virtual_column.substitute_expression.format(
        **virtual_column_replacements
      )
      tree = ast.parse(virtual_column_expression, mode='eval')
      valid = all(
        isinstance(node, query_editor.VALID_VIRTUAL_COLUMN_OPERATORS)
        for node in ast.walk(tree)
      )
      if valid:
        result = eval(
          compile(tree, filename='', mode='eval'), {'__builtins__': None}
        )
    except TypeError as e:
      raise exceptions.GaarfVirtualColumnException(
        f'cannot parse virtual_column {virtual_column.value}'
      ) from e
    except ZeroDivisionError:
      return 0
    except SyntaxError:
      return virtual_column.value
    return result


class ResourceFormatter:
  """Helper class for formatting resources strings."""

  def __init__(self, element: str) -> None:
    """Initializes ResourceFormatter based on element."""
    self.element = str(element).strip()

  def get_nested_resource(self) -> Self:
    """Extract nested resources from the API response field."""
    self.element = re.split(': ', self.element)[1]
    return self

  def get_resource_id(self) -> Self:
    """Extracts last id of resource_name.

    Resource name looks like `customer/123/campaigns/321`.
    `get_resource_id` returns `321`.
    """
    self.element = re.split('/', self.element)[-1]
    return self

  def clean_resource_id(self) -> Self:
    """Ensures that resource_id is cleaned up and converted to int."""
    self.element = re.sub('"', '', self.element)
    with contextlib.suppress(ValueError):
      self.element = int(self.element)
    return self

  def format(self) -> str | int:
    """Final method to return formatted resource.

    Returns:
      Formatted resource.
    """
    return self.element



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for executing queries in BigQuery."""

from __future__ import annotations

import warnings

from gaarf.executors import bq_executor


class BigQueryExecutor:
  """Deprecated class for creating BigQueryExecutor."""

  def __new__(
    cls, project_id: str, location: str | None = None
  ) -> bq_executor.BigQueryExecutor:
    warnings.warn(
      'Loading BigQueryExecutor from `gaarf.bq_executor` is '
      'deprecated; Import BigQueryExecutor from '
      '`gaarf.executors.bq_executor` instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return bq_executor.BigQueryExecutor(project_id, location)


def extract_datasets(macros: dict[str, str] | None = None) -> list[str]:
  """Deprecated method for extracting BigQuery datasets."""
  warnings.warn(
    'Loading `extract_datasets` from `gaarf.bq_executor` is '
    'deprecated; Import `extract_datasets` from '
    '`gaarf.executors.bq_executor` instead',
    category=DeprecationWarning,
    stacklevel=2,
  )
  return bq_executor.extract_datasets(macros)



from __future__ import annotations

import pytest

import gaarf.io.reader as reader


def test_console_reader():
  console_reader = reader.ConsoleReader()
  expected = 'SELECT 1'
  assert expected == console_reader.read(expected)


@pytest.fixture
def reader_factory():
  return reader.ReaderFactory()


def test_reader_factory_load(reader_factory):
  assert reader_factory.reader_options == {
    'file': reader.FileReader,
    'console': reader.ConsoleReader,
  }


def test_reader_factory_inits(reader_factory):
  file_reader = reader_factory.create_reader('file')
  console_reader = reader_factory.create_reader('console')
  assert isinstance(file_reader, reader.FileReader)
  assert isinstance(console_reader, reader.ConsoleReader)


def test_null_reader_raises_unknown_reader_error(reader_factory):
  with pytest.raises(ValueError):
    reader_factory.create_reader('non-existing-option')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data with CSV."""

from __future__ import annotations

import csv
import logging
import os
from typing import Literal

import smart_open

from gaarf.io import formatter
from gaarf.io.writers import file_writer
from gaarf.report import GaarfReport


class CsvWriter(file_writer.FileWriter):
  """Writes Gaarf Report to CSV.

  Attributes:
      destination_folder: Destination where CSV files are stored.
      delimiter: CSV delimiter.
      quotechar: CSV writer quotechar.
      quoting: CSV writer quoting.
  """

  def __init__(
    self,
    destination_folder: str | os.PathLike = os.getcwd(),
    delimiter: str = ',',
    quotechar: str = '"',
    quoting: Literal[0] = csv.QUOTE_MINIMAL,
    **kwargs,
  ) -> None:
    """Initializes CsvWriter based on a destination_folder.

    Args:
      destination_folder: Destination where CSV files are stored.
      delimiter: CSV delimiter.
      quotechar: CSV writer quotechar.
      quoting: CSV writer quoting.
      kwargs: Optional keyword arguments to initialize writer.
    """
    super().__init__(destination_folder=destination_folder, **kwargs)
    self.delimiter = delimiter
    self.quotechar = quotechar
    self.quoting = quoting

  def __str__(self):
    return (
      f'[CSV] - data are saved to {self.destination_folder} '
      'destination_folder.'
    )

  def write(self, report: GaarfReport, destination: str) -> str:
    """Writes Gaarf report to a CSV file.

    Args:
        report: Gaarf report.
        destination: Base file name report should be written to.

    Returns:
        Full path where data are written.
    """
    report = self.format_for_write(report)
    destination = formatter.format_extension(destination, new_extension='.csv')
    self.create_dir()
    logging.debug('Writing %d rows of data to %s', len(report), destination)
    output_path = os.path.join(self.destination_folder, destination)
    with smart_open.open(
      output_path,
      encoding='utf-8',
      mode='w',
    ) as file:
      writer = csv.writer(
        file,
        delimiter=self.delimiter,
        quotechar=self.quotechar,
        quoting=self.quoting,
      )
      writer.writerow(report.column_names)
      writer.writerows(report.results)
    logging.debug('Writing to %s is completed', output_path)
    return f'[CSV] - at {output_path}'
